/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.concurrent.locks.ReentrantLock;

import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;

/**
 * <p>
 * Abstract base class for {@link ChannelAccessMonitor} implementations. This
 * class takes care of handling the monitor listeners.
 * </p>
 * 
 * <p>
 * This class uses a {@link HashSet} to keep all
 * {@link ChannelAccessMonitorListener ChannelAccessMonitorListeners} that are
 * registered with this monitor. Derived classes should call the
 * {@link #notifyListeners(ChannelAccessGettableValue)} and
 * {@link #notifyListeners(ChannelAccessStatus, String)} methods in order to
 * notify the listeners. Implementations may choose to override the
 * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessGettableValue)}
 * and
 * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessStatus, String)}
 * methods in order to customize the handling of notifications (e.g. catch
 * certain exceptions possibly thrown by listeners).
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <ValueType>
 *            The exact type of the Channel Access values that are provided by
 *            the monitor events.
 */
public abstract class AbstractChannelAccessMonitor<ValueType extends ChannelAccessGettableValue<?>>
        implements ChannelAccessMonitor<ValueType> {

    /**
     * Registration of a monitor listener. This is a separate class because we
     * have to store some flags together with the lock object.
     * 
     * @author Sebastian Marsching
     *
     * @param <ValueType>
     *            value type of the monitor holding this listener registration..
     */
    private static class ListenerRegistration<ValueType extends ChannelAccessGettableValue<?>> {

        public final ChannelAccessMonitorListener<? super ValueType> listener;
        public boolean active = true;
        public boolean firstNotificationSent = false;
        public final ReentrantLock lock = new ReentrantLock();

        public ListenerRegistration(
                ChannelAccessMonitorListener<? super ValueType> listener) {
            this.listener = listener;
        }

    }

    private String lastEventErrorMessage;
    private ChannelAccessStatus lastEventErrorStatus;
    private ValueType lastEventValue;
    private boolean lastEventWasAnError;

    // We store the listener registrations in a map instead of a set. This way,
    // we can efficiently remove a registration for a specific listener.
    private HashMap<ChannelAccessMonitorListener<? super ValueType>, ListenerRegistration<ValueType>> listeners = new HashMap<ChannelAccessMonitorListener<? super ValueType>, ListenerRegistration<ValueType>>();

    /**
     * Synchronization object that this class synchronizes on when retrieving or
     * modifying the set of registered listeners or triggering the listeners.
     */
    protected final Object listenerLock = new Object();

    /**
     * Default constructor. This constructor is supposed to be used by classes
     * that inherit from this class. It initializes the monitor with an empty
     * set of listeners.
     */
    protected AbstractChannelAccessMonitor() {
    }

    /**
     * Notifies the listeners about an error. This method delegates calling each
     * listener to
     * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessStatus, String)}
     * so that derived classes can take extra precautions (e.g. catch certain
     * exceptions thrown by a listener). This method also saves the last error
     * so that listeners that are registered after the event can be notified
     * with the error status.
     * 
     * @param status
     *            Channel Access status code associated with the error.
     * @param message
     *            text message describing the error in greater detail (might be
     *            <code>null</code>).
     */
    protected void notifyListeners(ChannelAccessStatus status, String message) {
        ArrayList<ListenerRegistration<ValueType>> localListeners;
        ValueType localLastEventValue;
        synchronized (listenerLock) {
            lastEventWasAnError = true;
            lastEventErrorStatus = status;
            lastEventErrorMessage = message;
            localLastEventValue = lastEventValue;
            // We make a copy of the current set of listeners. This way, we do
            // not have to hold the connectionListenerLock when calling the
            // listeners.
            localListeners = new ArrayList<ListenerRegistration<ValueType>>(
                    listeners.values());
        }
        // Now, that we have released the listenerLock, we can notify the
        // listeners.
        for (ListenerRegistration<ValueType> listenerRegistration : localListeners) {
            // We have to synchronize on the per-listener lock while checking
            // the listener's registration and also keep the lock while
            // notifying the listener. This way, the listener cannot be removed
            // while we notify it and is never notified after it has been
            // removed.
            listenerRegistration.lock.lock();
            try {
                // We set the firstNotificationSet flag so that when the
                // listener has just been added, it does not receive an out of
                // order event. Without this flag, it could happen that the
                // listener is added, but before receiving its first event (with
                // stored data), a new event is triggered. In this case, the
                // event with the stored (older data) would be sent to the
                // listener after the newer event.
                if (listenerRegistration.active) {
                    // If the listener has never received a notification, we
                    // still notify it with the last-known "good" value, before
                    // sending the error event.
                    if (!listenerRegistration.firstNotificationSent
                            && localLastEventValue != null) {
                        listenerRegistration.firstNotificationSent = true;
                        notifyListener(listenerRegistration.listener,
                                localLastEventValue);
                    }
                    notifyListener(listenerRegistration.listener, status,
                            message);
                }
            } finally {
                listenerRegistration.lock.unlock();
            }
        }
    }

    /**
     * Notifies the listeners about a new value that was received. This method
     * delegates calling each listener to
     * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessGettableValue)}
     * so that derived classes can take extra precautions (e.g. catch certain
     * exceptions thrown by a listener). This method also saves the last value
     * so that listeners that are registered after the event can be notified
     * with the latest value.
     * 
     * @param value
     *            value that has been received from the server.
     */
    protected void notifyListeners(ValueType value) {
        ArrayList<ListenerRegistration<ValueType>> localListeners;
        synchronized (listenerLock) {
            lastEventWasAnError = false;
            lastEventValue = value;
            // We make a copy of the current set of listeners. This way, we do
            // not have to hold the connectionListenerLock when calling the
            // listeners.
            localListeners = new ArrayList<ListenerRegistration<ValueType>>(
                    listeners.values());
        }
        // Now, that we have released the listenerLock, we can notify the
        // listeners.
        for (ListenerRegistration<ValueType> listenerRegistration : localListeners) {
            // We have to synchronize on the per-listener lock while checking
            // the listener's registration and also keep the lock while
            // notifying the listener. This way, the listener cannot be removed
            // while we notify it and is never notified after it has been
            // removed.
            listenerRegistration.lock.lock();
            try {
                // We set the firstNotificationSet flag so that when the
                // listener has just been added, it does not receive an out of
                // order event. Without this flag, it could happen that the
                // listener is added, but before receiving its first event (with
                // stored data), a new event is triggered. In this case, the
                // event with the stored (older data) would be sent to the
                // listener after the newer event.
                listenerRegistration.firstNotificationSent = true;
                if (listenerRegistration.active) {
                    notifyListener(listenerRegistration.listener, value);
                }
            } finally {
                listenerRegistration.lock.unlock();
            }
        }
    }

    /**
     * Notifies a listener about an error. This method can be overridden in
     * order to modify how the listener is called. For example, some
     * implementations might want to catch certain exceptions thrown by a
     * listener.
     * 
     * @param listener
     *            the listener that should be notified. Notifications should
     *            happen by calling the listener's
     *            {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, ChannelAccessStatus, String)
     *            monitorError(...)} method.
     * @param status
     *            Channel Access status code associated with the error.
     * @param message
     *            text message describing the error in greater detail (might be
     *            <code>null</code>).
     */
    protected void notifyListener(
            ChannelAccessMonitorListener<? super ValueType> listener,
            ChannelAccessStatus status, String message) {
        listener.monitorError(this, status, message);
    }

    /**
     * Notifies a listener about a monitor event with a value that was received
     * from the server. This method can be overridden in order to modify how the
     * listener is called. For example, some implementations might want to catch
     * certain exceptions thrown by a listener.
     * 
     * @param listener
     *            the listener that should be notified. Notifications should
     *            happen by calling the listener's
     *            {@link ChannelAccessMonitorListener#monitorEvent(ChannelAccessMonitor, ChannelAccessGettableValue)
     *            monitorEvent(...)} method.
     * @param value
     *            value that has been received from the server.
     */
    protected void notifyListener(
            ChannelAccessMonitorListener<? super ValueType> listener,
            ValueType value) {
        listener.monitorEvent(this, value);
    }

    @Override
    public boolean addMonitorListener(
            ChannelAccessMonitorListener<? super ValueType> listener) {
        if (listener == null) {
            throw new NullPointerException(
                    "The listener parameter must not be null.");
        }
        ListenerRegistration<ValueType> listenerRegistration;
        ValueType localLastEventValue;
        boolean localLastEventWasAnError;
        ChannelAccessStatus localLastEventErrorStatus;
        String localLastEventErrorMessage;
        synchronized (listenerLock) {
            // If the listener has already been added, we do not do anything and
            // simply return.
            if (listeners.containsKey(listener)) {
                return false;
            }
            listenerRegistration = new ListenerRegistration<ValueType>(
                    listener);
            listeners.put(listener, listenerRegistration);
            localLastEventValue = lastEventValue;
            localLastEventWasAnError = lastEventWasAnError;
            localLastEventErrorStatus = lastEventErrorStatus;
            localLastEventErrorMessage = lastEventErrorMessage;
        }
        // We notify the listener with the last value that we received (if we
        // have one) and with the error status, if the last notification was an
        // error status. This gives the listener a chance to use the last good
        // value, even if there was an error in the meantime.
        // We do this after releasing the listenerLock. However, we acquire the
        // per-listener lock. This way, if a another thread removes the listener
        // while we are still in this method the listener will not get removed
        // before the notification has finished.
        // This also ensures that the listener will not receive the stored event
        // after it has received a newer event (triggered by in
        // notifyListeners(...).
        // If we cannot acquire the lock, we simply continue. This is safe
        // because there are only two reasons why the lock might be held by a
        // different thread: First, the listener might be in the process of
        // being removed. Second, the listener might be in the process of being
        // notified with a new event. In either case, there is no need for us to
        // call the listener now.
        boolean gotLock = listenerRegistration.lock.tryLock();
        if (gotLock) {
            try {
                // The listener might have been removed before we entered the
                // synchronized block or it might have been notified with a
                // newer event, so we have to check both flags.
                if (listenerRegistration.active
                        && !listenerRegistration.firstNotificationSent) {
                    listenerRegistration.firstNotificationSent = true;
                    if (localLastEventValue != null) {
                        notifyListener(listener, localLastEventValue);
                    }
                    if (localLastEventWasAnError) {
                        notifyListener(listener, localLastEventErrorStatus,
                                localLastEventErrorMessage);
                    }
                }
            } finally {
                listenerRegistration.lock.unlock();
            }
        }
        return true;
    }

    @Override
    public boolean removeMonitorListener(
            ChannelAccessMonitorListener<? super ValueType> listener) {
        ListenerRegistration<ValueType> listenerRegistration;
        synchronized (listenerLock) {
            listenerRegistration = listeners.remove(listener);
        }
        // We acquire the per-listener lock after releasing the listenerLock.
        // This way, we can avoid holding two locks at the same time.
        if (listenerRegistration != null) {
            listenerRegistration.lock.lock();
            try {
                listenerRegistration.active = false;
            } finally {
                listenerRegistration.lock.unlock();
            }
            return true;
        } else {
            return false;
        }
    }

}
