/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.lang.ref.WeakReference;
import java.util.LinkedList;

import org.apache.commons.lang3.StringEscapeUtils;

import com.aquenos.epics.jackie.client.AbstractChannelAccessChannel;
import com.aquenos.epics.jackie.client.ChannelAccessChannel;
import com.aquenos.epics.jackie.client.ChannelAccessClient;
import com.aquenos.epics.jackie.client.ChannelAccessConnectionListener;
import com.aquenos.epics.jackie.client.ChannelAccessMonitor;
import com.aquenos.epics.jackie.client.ClientThreadingStrategy;
import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.util.DelegatingListenableFuture;
import com.aquenos.epics.jackie.common.util.FutureCompletionListener;
import com.aquenos.epics.jackie.common.util.LinkedWeakIdentityHashSet;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Implementation of a {@link ChannelAccessChannel} that is used by the
 * {@link DefaultChannelAccessClient}. This class acts as a facade for a
 * {@link Channel} instance. This class is only intended for use by the classes
 * inside this package and by the {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * This class keeps a strong reference to its {@link ChannelAccessClient} so
 * that the client will not be destroyed as long as this channel instance
 * exists.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessChannelImpl extends AbstractChannelAccessChannel {

    /**
     * <p>
     * Custom listenable future implementation that delegates to another
     * listenable future but retains a strong reference to its channel instance.
     * </p>
     * 
     * <p>
     * We use this implementation as a wrapper around the future returned for
     * get and put operations. This way, it cannot happen that code starts such
     * an operation without further retaining a reference to the channel and
     * then the operation fails because its channel has been destroyed before it
     * finished.
     * </p>
     * 
     * <p>
     * We use a static class with an explicit reference to the channel for two
     * reasons: First, this allows us to release the reference when the future
     * has completed, avoiding the channel to be kept alive just because there
     * are some references to completed futures. Second, we do not have to worry
     * so much that a clever compiler detects that our class is in fact static
     * (because we do not use the implicit reference to the enclosing class) and
     * thus optimizes the reference away.
     * </p>
     * 
     * @author Sebastian Marsching
     *
     * @param <V>
     *            The result type returned by the future.
     */
    private static class InternalListenableFuture<V>
            extends DelegatingListenableFuture<V> {

        // Instead of using a simple listener implementation that accesses the
        // channel field directly, we only keep a weak reference to the
        // enclosing future. This way, the enclosing future can be garbage
        // collected if the operation has not finished yet but there are no
        // external references to the future any longer.
        private static class InternalListener<V>
                implements FutureCompletionListener<V> {

            private WeakReference<InternalListenableFuture<V>> internalFuture;

            public InternalListener(
                    InternalListenableFuture<V> internalFuture) {
                this.internalFuture = new WeakReference<InternalListenableFuture<V>>(
                        internalFuture);
            }

            @Override
            public void completed(ListenableFuture<? extends V> future) {
                // The future has completed, so the channel can now be destroyed
                // (if there are no other references to it).
                InternalListenableFuture<V> internalFuture = this.internalFuture
                        .get();
                if (internalFuture != null) {
                    internalFuture.channel = null;
                }
            }

        }

        // We keep a strong reference to the channel in order to keep it from
        // being garbage-collected before the future completes.
        @SuppressWarnings("unused")
        private ChannelAccessChannelImpl channel;

        public InternalListenableFuture(ChannelAccessChannelImpl channel,
                ListenableFuture<? extends V> delegate) {
            super(delegate);
            this.channel = channel;
            // After the future has completed, there is no reason to keep the
            // channel from being garbage-collected. Therefore, we register a
            // listener that is notified when the underlying future completes.
            // However, we make this listener only retain a weak reference to
            // this future, because it is okay to cancel the operation if there
            // are no external references to this future any longer.
            delegate.addCompletionListener(new InternalListener<V>(this));
        }

    }

    private Channel channel;
    private ChannelAccessClient client;

    // The destroyed flag or isDestroyed() method (which reads this flag) is
    // used in many places. By making destroyed volatile, we can avoid acquiring
    // the lock every time.
    private volatile boolean destroyed;
    private LinkedWeakIdentityHashSet<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>> monitors = new LinkedWeakIdentityHashSet<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>>();
    private final Object notificationLock = new Object();
    private LinkedWeakIdentityHashSet<ReadOperation<? extends ChannelAccessGettableValue<?>>> pendingReadOperations = new LinkedWeakIdentityHashSet<ReadOperation<? extends ChannelAccessGettableValue<?>>>();
    private LinkedWeakIdentityHashSet<WriteOperation> pendingWriteOperations = new LinkedWeakIdentityHashSet<WriteOperation>();
    private ClientThreadingStrategy threadingStrategy;

    /**
     * Creates a channel facade.
     * 
     * @param channelName
     *            name identifying the channel in the network.
     * @param client
     *            client that creates this channel and will "own" the channel.
     * @param channelPool
     *            channel pool for finding (or creating) a {@link Channel}
     *            instance that is going to be used by this channel facade.
     * @param threadingStrategy
     *            threading strategy used by the client. This channel instance
     *            uses the threading strategy when notifying connection
     *            listeners and monitor listeners.
     * @throws IllegalStateException
     *             if the resolver is already destroyed and thus no search
     *             request could be scheduled for the newly created channel.
     * @throws SerializedChannelNameTooLargeException
     *             if the serialized form of the channel name is too long to be
     *             sent in a search request.
     */
    public ChannelAccessChannelImpl(String channelName,
            ChannelAccessClient client, ChannelPool channelPool,
            ClientThreadingStrategy threadingStrategy) {
        this.client = client;
        this.threadingStrategy = threadingStrategy;
        while (this.channel == null) {
            // findOrCreateChannel(...) will throw an IllegalStateException if
            // the channel could not be started because the client is shutting
            // down and a SerializedChannelNameTooLargeException if the channel
            // name cannot be serialized. In both cases, we do not want to
            // handle the exceptions but forward them to the calling code.
            this.channel = channelPool.findOrCreateChannel(channelName);
            // We have to increment the use count so that the channel will not
            // be destroyed.
            try {
                this.channel.registerChannelFacade(this);
            } catch (IllegalStateException e) {
                // An IllegalStateException is triggered if the channel is
                // destroyed before we could register this facade. In this case,
                // we simply get a new instance and try to register again.
                this.channel = null;
            }
        }
        // We have to save the initial connection state. We only consider the
        // channel connected if the connection has been established and is
        // responsive. We acquire the mutex while updating the state to avoid a
        // race condition with the channel implementation calling one of our
        // event methods.
        synchronized (connectionStateLock) {
            // Calling the getState() method while holding the lock is not
            // perfect, but we do not have any choice. It should be okay though,
            // because the lock acquired by the getState() method should never
            // be hold when calling any of this class's methods.
            if (this.channel.getState().equals(Channel.State.CONNECTED)) {
                updateConnectionState(true);
            }
        }
    }

    /**
     * Returns the internal channel instance that is used by this facade. This
     * method is only intended for the {@link ChannelAccessMonitorImpl} class.
     * 
     * @return internal channel instance backing this facade.
     */
    public Channel getChannel() {
        return this.channel;
    }

    /**
     * Registers a monitor with this channel. Monitors are registered so that
     * they can be notified when this channel facade is destroyed. This method
     * is only intended for the {@link ChannelAccessMonitorImpl} class.
     * 
     * @param monitor
     *            monitor that has been created for this channel facade.
     */
    public void addMonitor(
            ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> monitor) {
        synchronized (connectionStateLock) {
            monitors.add(monitor);
        }
    }

    /**
     * Unregisters a monitor from this channels. Monitors are unregistered when
     * they are destroyed. This is done so that the internal set keeping (weak)
     * references to all monitors for this channel facade does not have to rely
     * on garbage collection. This way, the set does not grow unnecessarily
     * large when there is no garbage collection for a long time. This method is
     * only intended for the {@link ChannelAccessMonitorImpl} class.
     * 
     * @param monitor
     *            monitor that has previously been registered with this channel
     *            facade and has now been destroyed.
     */
    public void removeMonitor(
            ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> monitor) {
        synchronized (connectionStateLock) {
            monitors.remove(monitor);
        }
    }

    /**
     * Notifies the channel facade that the channel has been connected or become
     * responsive. This method is only intended for the {@link Channel} class.
     */
    void channelConnected() {
        // If this channel has been destroyed, we are not interested in any more
        // notifications. The listeners would not be called anyway.
        if (destroyed) {
            return;
        }
        // updateConnectionState(...) already handles redundant notifications
        // correctly, so there is no need for us to check the current connection
        // state before calling it.
        updateConnectionState(true);
    }

    /**
     * Notifies the channel facade that the channel has been disconnected or
     * become unresponsive. This method is only intended for the {@link Channel}
     * class.
     */
    void channelDisconnected() {
        // If this channel has been destroyed, we are not interested in any more
        // notifications. The listeners would not be called anyway.
        if (destroyed) {
            return;
        }
        // updateConnectionState(...) already handles redundant notifications
        // correctly, so there is no need for us to check the current connection
        // state before calling it.
        updateConnectionState(false);
    }

    @Override
    public void destroy() {
        // In order to set the destroyed state, we acquire both the
        // notificationLock and the connectionStateLock. The first lock ensures
        // that no connection listener is being notified after we have entered
        // the destroyed state. The second lock ensures that we do not interfere
        // with any other code that relies on the destroyed state.
        // The order in which we take the locks is important: A listener might
        // call a method of this channel which acquires the connectionStateLock.
        // As the listener already holds the notificationLock, there would be
        // the chance for a deadlock if we acquired the locks in the opposite
        // order.
        synchronized (notificationLock) {
            synchronized (connectionStateLock) {
                if (destroyed) {
                    return;
                }
                destroyed = true;
            }
        }
        // Cancel pending read and write operations. We do this because it might
        // be very irritating if operations finish after the corresponding
        // channel has been destroyed. We can access the sets without holding
        // the mutex because no other thread will touch the sets after this
        // channel has been marked as destroyed.
        for (ReadOperation<? extends ChannelAccessGettableValue<?>> operation : pendingReadOperations) {
            operation.operationCancelled();
        }
        pendingReadOperations.clear();
        for (WriteOperation operation : pendingWriteOperations) {
            operation.operationCancelled();
        }
        pendingWriteOperations.clear();
        // Destroy the monitors. We do this because it seems logical that there
        // cannot be any monitors for a channel that has been destroyed.
        // Technically, they still could exist if there is another channel
        // object referring to the same backing object, however this kind of
        // behavior would look erratic to the user. We can access the sets
        // without holding the mutex because no other thread will touch the sets
        // after this channel has been marked as destroyed.
        // We have to work on a copy of the monitors set because destroying a
        // monitor will remove it from the set, which would result in a
        // ConcurrentModificationException in the next iteration step.
        LinkedList<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>> monitors = new LinkedList<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>>(
                this.monitors);
        for (ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> monitor : monitors) {
            monitor.destroy();
        }
        // The set of monitors should now be empty because we destroyed all of
        // them.
        assert this.monitors.isEmpty();
        // We unregister the channel after releasing the mutex. This just looks
        // a little bit safer. We have to do this last anyway, because this
        // could lead to the backing channel object to be destroyed and we want
        // to destroy all monitors first.
        this.channel.unregisterChannelFacade(this);
    }

    @Override
    public String getName() {
        return channel.getName();
    }

    @Override
    public ChannelAccessClient getClient() {
        return client;
    }

    @Override
    public int getNativeCount() {
        // We check the destroyed flag while holding the lock but delegate to
        // the actual channel instance after releasing the lock. This helps to
        // avoid dead locks and the only risk is that for a channel that has
        // just been destroyed we still return the information instead of
        // throwing an exception. If the underlying channel has also been
        // destroyed or if it is disconnected, its method will throw an
        // IllegalStateException and we are fine.
        if (destroyed) {
            throw new IllegalStateException("This channel is destroyed.");
        }
        return channel.getNativeCount();
    }

    @Override
    public ChannelAccessValueType getNativeDataType() {
        // We check the destroyed flag while holding the lock but delegate to
        // the actual channel instance after releasing the lock. This helps to
        // avoid dead locks and the only risk is that for a channel that has
        // just been destroyed we still return the information instead of
        // throwing an exception. If the underlying channel has also been
        // destroyed or if it is disconnected, its method will throw an
        // IllegalStateException and we are fine.
        if (destroyed) {
            throw new IllegalStateException("This channel is destroyed.");
        }
        return channel.getNativeDataType();
    }

    @Override
    public boolean isConnected() {
        // The channel should not be considered connected if it has been
        // destroyed.
        return !destroyed && super.isConnected();
    }

    @Override
    public boolean isDestroyed() {
        return destroyed;
    }

    @Override
    public boolean isMayRead() {
        // We check the destroyed flag while holding the lock but delegate to
        // the actual channel instance after releasing the lock. This helps to
        // avoid dead locks and the only risk is that for a channel that has
        // just been destroyed we still return the flag instead of throwing an
        // exception. If the underlying channel has also been destroyed or if it
        // is disconnected, its method will throw an IllegalStateException and
        // we are fine.
        if (destroyed) {
            throw new IllegalStateException("This channel is destroyed.");
        }
        return channel.isMayRead();
    }

    @Override
    public boolean isMayWrite() {
        // We check the destroyed flag while holding the lock but delegate to
        // the actual channel instance after releasing the lock. This helps to
        // avoid dead locks and the only risk is that for a channel that has
        // just been destroyed we still return the flag instead of throwing an
        // exception. If the underlying channel has also been destroyed or if it
        // is disconnected, its method will throw an IllegalStateException and
        // we are fine.
        if (destroyed) {
            throw new IllegalStateException("This channel is destroyed.");
        }
        return channel.isMayWrite();
    }

    @Override
    public ListenableFuture<? extends ChannelAccessGettableValue<?>> get(
            ChannelAccessValueType valueType, int count) {
        if (count < 0) {
            throw new IllegalArgumentException(
                    "The count must not be negative.");
        }
        if (!valueType.isGettable()) {
            throw new IllegalArgumentException("The value type " + valueType
                    + " cannot be used with a get operation.");
        }
        if (destroyed) {
            throw new IllegalStateException("This channel is destroyed.");
        }
        // We can start the read operation without holding the mutex. If this
        // channel gets destroyed (or is disconnected) in the meantime, this
        // will be handled later.
        ReadOperation<? extends ChannelAccessGettableValue<?>> operation = channel
                .readValue(valueType, count);
        // Now we need the mutex again. We check whether this channel has been
        // destroyed and if not add the operation to the list of pending
        // operations so that it can be cancelled if this channel is destroyed.
        // The operation will never be removed explicitly, however we only keep
        // a weak reference, so eventually it will be garbage collected.
        synchronized (connectionStateLock) {
            if (destroyed) {
                operation.operationCancelled();
            } else {
                pendingReadOperations.add(operation);
            }
        }
        // We wrap the operation in a future that keeps a strong reference to
        // the channel. This way, if the calling code starts an operation
        // without further retaining a strong reference to the channel, the
        // operation will still complete instead of being cancelled because the
        // corresponding channel is destroyed. Basically, this is the same thing
        // that we are also doing for monitors.
        return new InternalListenableFuture<ChannelAccessGettableValue<?>>(this,
                operation);
    }

    @Override
    public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType, int count,
            ChannelAccessEventMask eventMask) {
        // We check the destroyed state before creating the monitor, so that we
        // can fail early.
        if (destroyed) {
            throw new IllegalStateException("This channel is destroyed.");
        }
        ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor = new ChannelAccessMonitorImpl<ChannelAccessGettableValue<?>>(
                this, valueType, count, eventMask, threadingStrategy);
        // After creating the monitor, we have to check the destroyed state
        // again, because it might have been created just after this channel got
        // destroyed. If this channel is destroyed later, the monitor will be
        // notified.
        if (destroyed) {
            monitor.destroy();
            throw new IllegalStateException("This channel is destroyed.");
        }
        return monitor;
    }

    @Override
    public ListenableFuture<Void> put(ChannelAccessPuttableValue<?> value) {
        if (value.getValueSize() == 0) {
            throw new IllegalArgumentException(
                    "The count must not be negative.");
        }
        if (destroyed) {
            throw new IllegalStateException("This channel is destroyed.");
        }
        // We can start the write operation without holding the mutex. If this
        // channel gets destroyed (or is disconnected) in the meantime, this
        // will be handled later.
        WriteOperation operation = channel.writeValue(value);
        // Now we need the mutex again. We check whether this channel has been
        // destroyed and if not add the operation to the list of pending
        // operations so that it can be cancelled if this channel is destroyed.
        // The operation will never be removed explicitly, however we only keep
        // a weak reference, so eventually it will be garbage collected.
        synchronized (connectionStateLock) {
            if (destroyed) {
                operation.operationCancelled();
            } else {
                pendingWriteOperations.add(operation);
            }
        }
        // We wrap the operation in a future that keeps a strong reference to
        // the channel. This way, if the calling code starts an operation
        // without further retaining a strong reference to the channel, the
        // operation will still complete instead of being cancelled because the
        // corresponding channel is destroyed. Basically, this is the same thing
        // that we are also doing for monitors.
        return new InternalListenableFuture<Void>(this, operation);
    }

    @Override
    protected void notifyConnectionListener(
            ChannelAccessConnectionListener listener, boolean nowConnected) {
        synchronized (notificationLock) {
            try {
                threadingStrategy.notifyConnectionListener(this, listener,
                        nowConnected);
            } catch (Throwable e) {
                client.getConfiguration().getErrorHandler().handleError(
                        this.getClass(), e,
                        "Connection listener for channel \""
                                + StringEscapeUtils
                                        .escapeJava(channel.getName())
                                + "\" threw an exception: " + e.getMessage());
            }
        }
    }

    @Override
    protected void notifyWaitForConnectionStateFutureListener(
            ListenableFuture<Boolean> future,
            FutureCompletionListener<? super Boolean> listener) {
        try {
            threadingStrategy.notifyFutureCompletionListener(future, listener);
        } catch (Throwable e) {
            client.getConfiguration().getErrorHandler().handleError(
                    this.getClass(), e,
                    "Future completion listener that was registered with a future for channel \""
                            + StringEscapeUtils.escapeJava(channel.getName())
                            + "\" threw an exception: " + e.getMessage());
        }
    }

    @Override
    protected void finalize() throws Throwable {
        try {
            // We have to call destroy in order to ensure that the underlying
            // channel's use counter is decremented (if this instance has not
            // been destroyed earlier).
            destroy();
        } finally {
            super.finalize();
        }
    }

}
