/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.util.LinkedList;
import java.util.concurrent.ExecutionException;

import org.apache.commons.lang3.tuple.Triple;

import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.util.FutureCompletionListener;
import com.aquenos.epics.jackie.common.util.IntegerIdPool;
import com.aquenos.epics.jackie.common.util.LinkedWeakIdentityHashSet;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Internal representation of the subscription to a Channel Access channel. This
 * class is only intended for use by the classes inside this package and by the
 * {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * Only one instance of this class is created for each channel and subscription
 * parameters (data type, count, event mask). An instance of this class lives as
 * long as there is at least one active {@link ChannelAccessMonitorImpl}
 * referring to it.
 * </p>
 * 
 * <p>
 * This implementation takes care of registering with the {@link Channel} and
 * acting on events received from the {@link Channel} or {@link Server},
 * forwarding them to its {@link ChannelAccessMonitorImpl} instances.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class Subscription {

    private Channel channel;
    private int count;
    private ChannelAccessValueType dataType;
    private boolean destroyed;
    private String lastErrorMessage;
    private ChannelAccessStatus lastErrorStatus;
    private ChannelAccessGettableValue<?> lastValue;
    private final Object lock = new Object();
    private ChannelAccessEventMask mask;
    private LinkedWeakIdentityHashSet<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>> monitorFacades = new LinkedWeakIdentityHashSet<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>>();
    private boolean noEventSinceServerResponsive;
    private IntegerIdPool.Id subscriptionId;

    /**
     * Creates a subscription for the specified channel, using the specified
     * parameters.
     * 
     * @param channel
     *            channel for which the subscription shall be created.
     * @param count
     *            maximum number of elements that the server shall send with
     *            event notifications. If zero, the channel's native element
     *            count is used.
     * @param dataType
     *            data type of the values that the server shall send with event
     *            notifications. If <code>null</code>, the channel's native
     *            data-type is used.
     * @param mask
     *            event mask used for the subscription. This must not be
     *            {@link ChannelAccessEventMask#DBE_NONE}.
     * @param subscriptionId
     *            client-assigned ID for the subscription. This ID must be
     *            unique for the whole client (across all channels).
     */
    public Subscription(Channel channel, int count,
            ChannelAccessValueType dataType, ChannelAccessEventMask mask,
            IntegerIdPool.Id subscriptionId) {
        // The mask should already have been checked by the calling
        // ChannelAccessMonitorImpl. Therefore, we only use an assertion here.
        assert mask.isAny();
        this.channel = channel;
        this.count = count;
        this.dataType = dataType;
        this.mask = mask;
        this.subscriptionId = subscriptionId;
    }

    /**
     * Returns the channel for which this subscription has been created.
     * 
     * @return channel of this subscription.
     */
    public Channel getChannel() {
        return channel;
    }

    /**
     * Returns the element count which has been specified when creating this
     * subscription. An element count of zero indicates that the channel's
     * native element count shall be used.
     * 
     * @return element count requested for this subscription.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the data type which has been specified when creating this
     * subscription. A data type of <code>null</code> indicates that the
     * channel's native data-type shall be used.
     * 
     * @return data type requested for this subscription or <code>null</code> if
     *         the channel's native data-type shall be used.
     */
    public ChannelAccessValueType getDataType() {
        return dataType;
    }

    /**
     * Returns the event mask that has been requested when creating this
     * subscription. The event mask defines when the server sends event
     * notifications. The event mask for a subscription cannot be
     * {@link ChannelAccessEventMask#DBE_NONE}.
     * 
     * @return event mask for this subscription.
     */
    public ChannelAccessEventMask getMask() {
        return mask;
    }

    /**
     * Returns the subscription ID that has been assigned to this subscription.
     * The subscription ID uniquely identifies the subscription within the whole
     * client.
     * 
     * @return subscription ID.
     */
    public IntegerIdPool.Id getSubscriptionId() {
        return subscriptionId;
    }

    /**
     * Tells whether this subscription has been destroyed.
     * 
     * @return <code>true</code> if the subscription has been destroyed,
     *         <code>false</code> if it is still alive.
     */
    public boolean isDestroyed() {
        synchronized (lock) {
            return destroyed;
        }
    }

    /**
     * <p>
     * Registers a monitor facade with this subscription. The monitor facade is
     * the object that the user-code sees as the monitor/subscription instance.
     * Internally, it delegates all actions to this subscription.
     * </p>
     * 
     * <p>
     * The list of registered monitor facades is internally kept through weak
     * references. This allows the monitor facades to be garbage collected when
     * they are not used any longer. When this list becomes empty (after at
     * least one monitor facade has been registered with this subscription),
     * this subscription is destroyed.
     * </p>
     * 
     * @param monitorFacade
     *            monitor facade that is going to use this subscription.
     * @return latest value and error status received from the server. The error
     *         status is only returned if it is newer than the latest value. If
     *         no events have been received yet, the value might be
     *         <code>null</code> as well.
     * @throws IllegalStateException
     *             if this subscription has been destroyed.
     * @see #unregisterMonitorFacade(ChannelAccessMonitorImpl)
     */
    public Triple<ChannelAccessGettableValue<?>, ChannelAccessStatus, String> registerMonitorFacade(
            ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> monitorFacade) {
        // We register and get the currently saved values in a single step while
        // holding the mutex. This ensures that the monitor facade will get all
        // events that change the state compared to the state returned by the
        // call to this method.
        synchronized (lock) {
            // We do not accept new registrations after the subscription has
            // been destroyed. A new subscription should be created instead.
            if (destroyed) {
                throw new IllegalStateException(
                        "This subscription is destroyed.");
            }
            monitorFacades.add(monitorFacade);
            // For some reason, we get a compiler error if we do not specify the
            // generic parameters to the of(...) method explicitly.
            return Triple
                    .<ChannelAccessGettableValue<?>, ChannelAccessStatus, String> of(
                            lastValue, lastErrorStatus, lastErrorMessage);
        }
    }

    /**
     * <p>
     * Unregisters a monitor facade from this channel. The monitor facade is the
     * object that the user-code sees as the monitor/subscription instance.
     * Internally, it delegates all actions to this subscription.
     * </p>
     * 
     * <p>
     * The list of registered monitor facades is internally kept through weak
     * references. Even though this means that this references will eventually
     * be cleaned up if the monitor facade is destroyed, we still unregister
     * monitor facades explicitly to allow for an early release of resources.
     * This is also important to make sure that the subscription is destroyed
     * when there are no more monitor facades registered to it.
     * </p>
     * 
     * @param channelAccessMonitorImpl
     *            monitor facade that has been registered with this subscription
     *            and is not using this subscription any longer.
     * @see #registerMonitorFacade(ChannelAccessMonitorImpl)
     */
    public void unregisterMonitorFacade(
            ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> channelAccessMonitorImpl) {
        // As notifications happen without holding the mutex, it is possible
        // that a monitor facade receives event notifications after it has
        // unregistered. However, a monitor facade only unregisters when it is
        // destroyed and it simply ignores notifications that it received after
        // having been destroyed.
        boolean destroyedInternal;
        synchronized (lock) {
            // We can assume that this method is never called after the
            // subscription has been destroyed. It is only destroyed when all
            // monitor facades have unregistered.
            assert !destroyed;
            monitorFacades.remove(channelAccessMonitorImpl);
            // We want to destroy this subscription when the last monitor facade
            // is removed.
            if (monitorFacades.isEmpty()) {
                destroyed = true;
            }
            destroyedInternal = destroyed;
        }
        // Destroyed is only set once, therefore we can take care of
        // unregistering the subscription after releasing the mutex.
        if (destroyedInternal) {
            channel.removeSubscription(this);
        }
    }

    /**
     * Notifies the subscription that it has been cancelled by the server.
     * Typically, this happens after the subscription has been destroyed. This
     * method may only be called by the {@link Server} associated with this
     * subscription.
     * 
     * @param server
     *            server sending the notification. This can be used to detect
     *            that a notification is from an old server when the channel is
     *            now connected with a different server or has disconnected.
     */
    void cancelled(Server server) {
        synchronized (lock) {
            // We check whether the server that sends the notification is the
            // same server that is currently used by our channel. This makes
            // sure that we will not process events that arrived after the
            // channel unregistered with the server (and potentially registered
            // with a different server). It looks like there is a race condition
            // here because the server could change after we check it, but the
            // channel's server only changes if the channel disconnects from a
            // server or is destroyed. The first thing only happens in the
            // communication thread (which is the same thread that calls this
            // method) and the second case is not a problem because all
            // subscriptions for a channel are destroyed before the channel is
            // destroyed (indirectly - all monitor facades are destroyed which
            // will also destroy the subscription if the last channel facade has
            // been destroyed and thus there are not any other monitor facades.
            // We can do this check while holding the mutex because this mutex
            // and the server's mutex are always acquired in this order. And
            // thus there is no risk of a dead lock.
            if (channel.getServer() != server) {
                return;
            }
            // We should only receive a cancelled notification if this
            // subscription has been destroyed, but it is still better to check
            // because a server might send such a message erroneously.
            if (destroyed) {
                // Now that the subscription has also been cancelled on the
                // server, we can reuse the subscription ID.
                subscriptionId.release();
            }
        }
        // Finally, we notify the channel that the subscription has been
        // cancelled. This allows the server to remove this channel from the
        // list of subscriptions that are pending cancellation.
        channel.subscriptionCancelled(this);
    }

    /**
     * Notifies the subscription that an error regarding the subscription has
     * been received from the server. This method may only be called by the
     * {@link Channel} and {@link Server} associated with this subscription.
     * 
     * @param server
     *            server sending the notification. This can be used to detect
     *            that a notification is from an old server when the channel is
     *            now connected with a different server or has disconnected.
     * @param status
     *            Channel Access status associated with the error.
     * @param errorMessage
     *            textual representation of the error or <code>null</code> if
     *            the server did not send such a message with the error.
     */
    void errorReceived(Server server, ChannelAccessStatus status,
            String errorMessage) {
        LinkedList<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>> listeners;
        synchronized (lock) {
            // We check whether the server that sends the notification is the
            // same server that is currently used by our channel. This makes
            // sure that we will not process events that arrive after the
            // channel unregistered from the server (and potentially registered
            // with a different server). It looks like there is a race condition
            // here because the server could change after we check it, but the
            // channel's server only changes if the channel disconnects from a
            // server or is destroyed. The first thing only happens in the
            // communication thread (which is the same thread that calls this
            // method) and the second case is not a problem because all
            // subscriptions for a channel are destroyed before the channel is
            // destroyed (indirectly - all monitor facades are destroyed which
            // will also destroy the subscription if the last channel facade has
            // been destroyed and thus there are no other monitor facades.
            // We can do this check while holding the mutex because this mutex
            // and the server's mutex are always acquired in this order. And
            // thus there is no risk of a dead lock.
            if (channel.getServer() != server) {
                return;
            }
            if (destroyed) {
                return;
            }
            // We save the error status and message so that we can return it to
            // newly registered listeners.
            lastErrorStatus = status;
            lastErrorMessage = errorMessage;
            // If we got an event since an unresponsive server became
            // responsive, we do not want to use a value from an explicit read
            // operation. Therefore, we set the corresponding flag to false.
            noEventSinceServerResponsive = false;
            // We have to copy the list of listeners while holding the mutex. If
            // a listener is registered after we release the mutex, it also gets
            // the latest error while registering, so it does not matter that it
            // is not notified from here.
            listeners = new LinkedList<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>>(
                    monitorFacades);
        }
        // We do the actual notifications after releasing the mutex. This helps
        // to avoid a dead lock because the listener will acquire a mutex of
        // its own. There is no race condition affecting the order of
        // notifications because all notifications are sent inside the
        // communication thread.
        for (ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> listener : listeners) {
            listener.errorReceived(status, errorMessage);
        }
    }

    /**
     * Notifies the subscription that a subscription event has been received
     * from the server. This method may only be called by the {@link Server}
     * associated with this subscription.
     * 
     * @param server
     *            server sending the notification. This can be used to detect
     *            that a notification is from an old server when the channel is
     *            now connected with a different server or has disconnected.
     * @param status
     *            Channel Access status associated with the event. If the status
     *            message is not
     *            {@link ChannelAccessStatus.StatusMessage#ECA_NORMAL}, the
     *            value is invalid (or <code>null</code>) and an error occurred
     *            on the server while trying to read the channel's value.
     * @param value
     *            value read from the channel and associated with the event.
     *            Only valid if the <code>status</code> indicates success.
     */
    void eventReceived(Server server, ChannelAccessStatus status,
            ChannelAccessGettableValue<?> value) {
        LinkedList<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>> listeners;
        boolean isError;
        synchronized (lock) {
            // We check whether the server that sends the notification is the
            // same server that is currently used by our channel. This makes
            // sure that we will not process events that arrive after the
            // channel unregistered from the server (and potentially registered
            // with a different server). It looks like there is a race condition
            // here because the server could change after we check it, but the
            // channel's server only changes if the channel disconnects from a
            // server or is destroyed. The first thing only happens in the
            // communication thread (which is the same thread that calls this
            // method) and the second case is not a problem because all
            // subscriptions for a channel are destroyed before the channel is
            // destroyed (indirectly - all monitor facades are destroyed which
            // will also destroy the subscription if the last channel facade has
            // been destroyed and thus there are no other monitor facades.
            // We can do this check while holding the mutex because this mutex
            // and the server's mutex are always acquired in this order. And
            // thus there is no risk of a dead lock.
            if (channel.getServer() != server) {
                return;
            }
            if (destroyed) {
                return;
            }
            // If the status is OK, we use the value, otherwise we do not use it
            // and signal an error condition instead. We save the value and
            // error information so that we can return it to newly registered
            // listeners.
            if (status
                    .matchesMessage(ChannelAccessStatus.StatusMessage.ECA_NORMAL)) {
                // The monitor event indicates success. However, we need to
                // verify that the data type and count match the request.
                if ((dataType != null && !value.getType().equals(dataType))
                        || (value.getValueSize() > count && count != 0)) {
                    isError = true;
                    lastErrorStatus = ChannelAccessStatus.StatusMessage.ECA_GETFAIL
                            .toStatus();
                    lastErrorMessage = null;
                } else {
                    isError = false;
                    lastValue = value;
                    lastErrorStatus = null;
                    lastErrorMessage = null;
                }
            } else {
                isError = true;
                lastErrorStatus = status;
                lastErrorMessage = null;
            }
            // If we got an event since an unresponsive server became
            // responsive, we do not want to use a value from an explicit read
            // operation. Therefore, we set the corresponding flag to false.
            noEventSinceServerResponsive = false;
            // We have to copy the list of listeners while holding the mutex. If
            // a listener is registered after we release the mutex, it also gets
            // the latest value while registering, so it does not matter that it
            // is not notified from here.
            listeners = new LinkedList<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>>(
                    monitorFacades);
        }
        // We do the actual notifications after releasing the mutex. This helps
        // to avoid a dead lock because the listener will acquire a mutex of
        // its own. There is no race condition affecting the order of
        // notifications because all notifications are sent inside the
        // communication thread.
        if (isError) {
            for (ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> listener : listeners) {
                listener.errorReceived(status, null);
            }
        } else {
            for (ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> listener : listeners) {
                listener.eventReceived(value);
            }
        }
    }

    /**
     * Notifies the subscription that the server which is used by its channel
     * has disconnected. This kind of notification is only sent if the
     * subscription has been destroyed and has requested to be cancelled. As a
     * subscription is cancelled automatically when the connection is lost, the
     * subscription does not have to wait to be cancelled any longer. This
     * method may only be called by the {@link Channel} associated with this
     * subscription.
     */
    void serverDisconnected() {
        synchronized (lock) {
            // We can assume that the channel only notifies us about a server
            // disconnect after we requested removal of this subscription, so
            // this subscription must have been destroyed.
            assert destroyed;
            // Now that the server disconnected, there is no risk that it might
            // still refer to the subscription ID and we can reuse it.
            subscriptionId.release();
        }
    }

    /**
     * Notifies the subscription that the server for its channel has become
     * responsive. The subscription uses this information to schedule a read
     * request with the channel, because it might have lost subscription events
     * while the connection was unresponsive. This method may only be called by
     * the {@link Server} associated with this subscription.
     * 
     * @param server
     *            server sending the notification. This can be used to detect
     *            that a notification is from an old server when the channel is
     *            now connected with a different server or has disconnected.
     */
    void serverResponsive(Server server) {
        final ReadOperation<? extends ChannelAccessGettableValue<?>> readOperation;
        synchronized (lock) {
            try {
                readOperation = channel.readValue(dataType, count);
            } catch (IllegalStateException e) {
                // If readValue(...) throws, the channel is not connected.
                // In this case, we will get a new event when the channel
                // reconnects, so we do not have to do anything.
                return;
            }
            noEventSinceServerResponsive = true;
        }
        // We register the listener after releasing the mutex, because the
        // listener might get executed right on registration.
        readOperation
                .addCompletionListener(new FutureCompletionListener<ChannelAccessGettableValue<?>>() {

                    @Override
                    public void completed(
                            ListenableFuture<? extends ChannelAccessGettableValue<?>> future) {
                        ChannelAccessGettableValue<?> value = null;
                        ChannelAccessStatus status = null;
                        String errorMessage = null;
                        LinkedList<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>> listeners;
                        synchronized (lock) {
                            // If this subscription got destroyed in the
                            // meantime, we do not want to use the data.
                            if (destroyed) {
                                return;
                            }
                            // If we received a regular update in the meantime,
                            // we do not want to use the value from the read
                            // request.
                            if (!noEventSinceServerResponsive) {
                                return;
                            }
                            try {
                                value = readOperation.get();
                                lastValue = value;
                                lastErrorStatus = null;
                                lastErrorMessage = null;
                            } catch (InterruptedException e) {
                                // We do not have to wait on the result because
                                // we know that the operation has finished, so
                                // we should never get an InterruptedException.
                                throw new AssertionError(
                                        "Unexpected InterruptedException from future that should have been finished.");
                            } catch (ExecutionException e) {
                                Throwable cause = e.getCause();
                                // Any execution exception that we throw from a
                                // future must have a cause.
                                assert (cause != null);
                                if (cause instanceof ChannelAccessException) {
                                    ChannelAccessException caException = (ChannelAccessException) cause;
                                    status = caException.getStatus();
                                    errorMessage = caException
                                            .getExtendedMessage();
                                    lastErrorStatus = status;
                                    lastErrorMessage = errorMessage;
                                } else {
                                    // There might be other exceptions, e.g. if
                                    // the server disconnects while the read
                                    // operation is pending. We ignore all those
                                    // exceptions and simply return because
                                    // there is no useful way we could act on
                                    // them. Typically this is okay, because
                                    // eventually, the subscription will get
                                    // registered again and we will start
                                    // receiving new events.
                                    return;
                                }
                            }
                            listeners = new LinkedList<ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>>>(
                                    monitorFacades);
                        }
                        // After releasing the mutex, we notify our listeners.
                        // The fact that this method is only called within the
                        // communication thread (unless the operation has been
                        // cancelled which is okay because we return early)
                        // ensures the right ordering.
                        if (status != null) {
                            for (ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> monitorFacade : listeners) {
                                monitorFacade.errorReceived(status,
                                        errorMessage);
                            }
                        } else if (value != null) {
                            for (ChannelAccessMonitorImpl<? extends ChannelAccessGettableValue<?>> monitorFacade : listeners) {
                                monitorFacade.eventReceived(value);
                            }
                        }
                    }
                });
    }

}
