/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.exception;

/**
 * <p>
 * Wrapping error handler that ensures that no exceptions are thrown. This
 * wrapper can be used by code that calls an error handler to ensure that the
 * {@link #handleError(Class, Throwable, String)} method does not throw an
 * exception, if there is an implementation bug in the error handler.
 * </p>
 * 
 * <p>
 * This error handler simply delegates to the error handler specified at
 * construction time and catches any exception the delegate error-handler might
 * throw.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ExceptionSafeErrorHandler implements ErrorHandler {

    private ErrorHandler delegate;

    /**
     * Creates an error handler that never thrown an exception from the
     * specified error handler. If the specified error handler is
     * <code>null</code>, this error handler simply ignores all calls to
     * {@link #handleError(Class, Throwable, String)}.
     * 
     * @param delegate
     *            error handler to which calls of
     *            {@link #handleError(Class, Throwable, String)} should be
     *            delegated or <code>null</code> if these calls should simply be
     *            ignored.
     */
    public ExceptionSafeErrorHandler(ErrorHandler delegate) {
        this.delegate = delegate;
    }

    @Override
    public void handleError(Class<?> context, Throwable e, String description) {
        if (delegate == null) {
            return;
        }
        try {
            delegate.handleError(context, e, description);
        } catch (Throwable throwable) {
            // We simply ignore any exception thrown by the error handler.
        }
    }

}
