/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_EVENTS_ON</code> message. This message is
 * sent from a client to a server via TCP in order to indicate that the server
 * should resume sending events for all registered monitors after the events
 * have been paused earlier by sending a events can later be resumed by sending
 * the {@link ChannelAccessEventsOffMessage}.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_EVENTS_ON
 */
public final class ChannelAccessEventsOnMessage extends ChannelAccessMessage {

    /**
     * Constructor for creating a <code>CA_PROTO_EVENTS_ON</code> message.
     */
    public ChannelAccessEventsOnMessage() {
        super(ChannelAccessCommand.CA_PROTO_EVENTS_ON);
    }

    private ChannelAccessEventsOnMessage(
            ChannelAccessMessageHeader messageHeader) {
        super(ChannelAccessCommand.CA_PROTO_EVENTS_ON, messageHeader);
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, (short) 0, 0, 0, 0, version,
                maxPayloadSize);
    }

    // This message has no state, so there is no need to override equals and
    // hashCode from the super class.

    @Override
    public String toString() {
        return command.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including the payload).
     */
    protected static ChannelAccessEventsOnMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_EVENTS_ON
                .getCommandNumber());
        if (!headerOnly) {
            byteSource.skip(messageHeader.getPayloadSize());
        }
        return new ChannelAccessEventsOnMessage(messageHeader);
    }

}
