/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.UnsupportedProtocolVersionException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.Inet4AddressUtil;

/**
 * Representation of a <code>CA_PROTO_SEARCH</code> message that is sent from a
 * server to a client via TCP in reply to a
 * {@link ChannelAccessSearchClientMessage}. This message carries information
 * about how to connect to the requested channel.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_SEARCH
 */
public final class ChannelAccessSearchTCPServerMessage extends
        ChannelAccessSearchMessage {

    private final static int UNKNOWN_SERVER_ADDRESS = (int) 0xffffffff;

    private int cid;
    private Inet4Address serverAddress;
    private int serverPort;

    /**
     * Creates a <code>CA_PROTO_SEARCH</code> message that can be sent from a
     * server to a client via TCP.
     * 
     * @param cid
     *            client-side channel identifier. This is the identifier that
     *            has been sent by the client with the search request and is
     *            used by the client to map the response to the right channel.
     * @param serverAddress
     *            IP address of the server serving the channel. If this is
     *            <code>null</code> or the wildcard address, the client will try
     *            to connect to the server sending this message.
     * @param serverPort
     *            TCP port of the server serving the channel. The port number is
     *            only used if <code>serverAddress</code> specifies a valid IP
     *            address (not <code>null</code> or the wildcard address).
     *            Otherwise, the client will try to connect to the port of the
     *            server that sent this message.
     */
    public ChannelAccessSearchTCPServerMessage(int cid,
            Inet4Address serverAddress, int serverPort) {
        super();

        this.cid = cid;
        this.serverAddress = serverAddress;
        this.serverPort = serverPort;
    }

    private ChannelAccessSearchTCPServerMessage(
            ChannelAccessMessageHeader messageHeader, int cid,
            Inet4Address serverAddress, int serverPort) {
        super(messageHeader);

        this.cid = cid;
        this.serverAddress = serverAddress;
        this.serverPort = serverPort;
    }

    /**
     * Returns the channel ID that is used by the client to identify the
     * channel. This integer may be zero or negative.
     * 
     * @return client-side ID for the channel.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the IP address of the server that is serving the requested
     * channel.
     * 
     * @return IP address of the server for the channel.
     */
    public Inet4Address getServerAddress() {
        return serverAddress;
    }

    /**
     * Returns the TCP port number on which the server that is serving the
     * requested channel is listening.
     * 
     * @return TCP port number.
     */
    public int getServerPort() {
        return serverPort;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new UnsupportedProtocolVersionException(
                    "This library does not support clients with a protocol version before Channel Access version "
                            + ChannelAccessVersion.V4_4 + ".");
        }
        int serverAddress;
        short port;
        if (this.serverAddress != null
                && !this.serverAddress.isAnyLocalAddress()) {
            serverAddress = Inet4AddressUtil.addressToInt(this.serverAddress
                    .getAddress());
        } else {
            serverAddress = UNKNOWN_SERVER_ADDRESS;
        }
        port = (short) this.serverPort;
        serializeHeader(byteSink, 0, port, 0, serverAddress, cid, version,
                maxPayloadSize);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new UnsupportedProtocolVersionException(
                    "This library does not support clients with a protocol version before Channel Access version "
                            + ChannelAccessVersion.V4_4 + ".");
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessSearchTCPServerMessage other = (ChannelAccessSearchTCPServerMessage) obj;
        return new EqualsBuilder().append(this.cid, other.cid)
                .append(this.serverAddress, other.serverAddress)
                .append(this.serverPort, other.serverPort).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .append(serverAddress).append(serverPort).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append(", serverAddress=");
        sb.append(serverAddress != null ? serverAddress.getHostAddress()
                : "null");
        sb.append(", serverPort=");
        sb.append(serverPort);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param originatingServerAddress
     *            IP address of the server that sent this message.
     * @param originatingServerPort
     *            UDP port of the server that sent this message.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     */
    protected static ChannelAccessSearchTCPServerMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            Inet4Address originatingServerAddress, int originatingServerPort) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_SEARCH
                .getCommandNumber());
        byteSource.skip(messageHeader.getPayloadSize());
        Inet4Address serverAddress;
        int serverPort;
        int serverAddressAsInt = messageHeader.getCID();
        if (serverAddressAsInt != UNKNOWN_SERVER_ADDRESS) {
            InetAddress tempAddress;
            try {
                tempAddress = InetAddress.getByAddress(Inet4AddressUtil
                        .addressToBytes(serverAddressAsInt));
            } catch (UnknownHostException e) {
                throw new RuntimeException(
                        "Cannot get inet address instance for address of four bytes.",
                        e);
            }
            if (!(tempAddress instanceof Inet4Address)) {
                throw new RuntimeException(
                        "Did not get IPv4 address where it was expected. Got address of type "
                                + tempAddress.getClass().getName()
                                + " instead.");
            }
            serverAddress = (Inet4Address) tempAddress;
            serverPort = messageHeader.getDataType() & 0xffff;
        } else {
            serverAddress = originatingServerAddress;
            serverPort = originatingServerPort;
        }
        return new ChannelAccessSearchTCPServerMessage(messageHeader,
                messageHeader.getContextSpecific(), serverAddress, serverPort);
    }

}
