/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.ShortPayloadException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Representation of a <code>CA_PROTO_EVENT_ADD</code> message that is sent from
 * a client to a server. This message is sent in order to subscribe to events
 * for a channel.
 * 
 * @see ChannelAccessCommand#CA_PROTO_EVENT_ADD
 * @see ChannelAccessSubscriptionServerMessage
 * @see ChannelAccessCancelSubscriptionServerMessage
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessSubscriptionClientMessage extends
        ChannelAccessSubscriptionMessage {

    private static final byte[] TWELVE_ZERO_BYTES = new byte[12];
    private static final int EXPECTED_PAYLOAD_SIZE = 16;

    private short dataType;
    private int count;
    private int sid;
    private int subscriptionId;
    private short mask;

    /**
     * Creates a <code>CA_PROTO_EVENT_ADD</code> message. This constructor
     * should be used to create a message that is supposed to be sent from a
     * client to a server.
     * 
     * @param dataType
     *            requested data-type for the subscription. The server should
     *            send notifications with this data type.
     * @param count
     *            the number of value elements to send in notifications. Must
     *            not be larger than the channel's native element count. May
     *            only be zero for Channel Access version 4.13 and newer. In
     *            this case, a value of zero indicates to use the channel's
     *            native count.
     * @param sid
     *            server-assigned ID identifying the channel.
     * @param subscriptionId
     *            client-assigned ID identifying the subscription. This ID will
     *            be sent with events so that the client can identify the
     *            subscription to which the event belongs.
     * @param mask
     *            event mask to use for the subscription. A subscription will
     *            only receive events which have an event mask that has at least
     *            one bit set, that it also set in the subscription's event mask
     *            (binary or).
     */
    public ChannelAccessSubscriptionClientMessage(
            ChannelAccessValueType dataType, int count, int sid,
            int subscriptionId, ChannelAccessEventMask mask) {
        super();

        this.dataType = dataType.toTypeCode();
        this.count = count;
        this.sid = sid;
        this.subscriptionId = subscriptionId;
        this.mask = mask.toNumericMask();
    }

    private ChannelAccessSubscriptionClientMessage(
            ChannelAccessMessageHeader messageHeader, short dataType,
            int count, int channelSID, int subscriptionId, short mask) {
        super(messageHeader);

        this.dataType = dataType;
        this.count = count;
        this.sid = channelSID;
        this.subscriptionId = subscriptionId;
        this.mask = mask;
    }

    /**
     * Returns the data-type requested for the subscription. This is the data
     * type that the server should use when sending notifications for this
     * subscription. Returns <code>null</code> if the specified data-type is
     * unknown to this implementation.
     * 
     * @return data-type for the subscription or <code>null</code> if the
     *         numeric identifier cannot be mapped to a data type known by this
     *         implementation.
     */
    public ChannelAccessValueType getDataType() {
        try {
            return ChannelAccessValueType.forTypeNumber(dataType);
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * Returns the number of elements of the value that should be sent in
     * subscription events. This number should not exceed the native count sent
     * by the server when the channel was connected. A count of zero is only
     * allowed since Channel Access version 4.13. Since this version, this means
     * that the server should use the native count of the channel.
     * 
     * @return number of value elements requested by the client.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the channel SID. This is the identifier that the server uses to
     * identify the channel. This number might be negative or zero.
     * 
     * @return channel SID assigned to the channel by the server.
     */
    public int getChannelSID() {
        return sid;
    }

    /**
     * Returns the client-assigned subscription ID. The subscription ID is used
     * by the client in order to map events sent in response to a subscription
     * to the respective subscription. This number might be negative or zero.
     * 
     * @return client-assigned subscription identifier.
     */
    public int getSubscriptionId() {
        return subscriptionId;
    }

    /**
     * Returns the event mask for this subscription. The event mask defines
     * which kind of events should be sent for this subscription.
     * 
     * @return event mask for this subscription request.
     */
    public ChannelAccessEventMask getMask() {
        return ChannelAccessEventMask.forNumericMask(mask);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessSubscriptionClientMessage other = (ChannelAccessSubscriptionClientMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count).append(this.sid, other.sid)
                .append(this.subscriptionId, other.subscriptionId)
                .append(this.mask, other.mask).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(sid)
                .append(subscriptionId).append(mask).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append(", sid=");
        sb.append(sid);
        sb.append(", subscriptionId=");
        sb.append(subscriptionId);
        sb.append(", mask=");
        sb.append(ChannelAccessEventMask.forNumericMask(mask));
        sb.append('}');
        return sb.toString();
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        // The payload always has exactly 16 bytes, but only two of them are
        // used.
        serializeHeader(byteSink, 16, dataType, count, sid, subscriptionId,
                version, maxPayloadSize);
        byteSink.putByteArray(TWELVE_ZERO_BYTES);
        byteSink.putShort(mask);
        byteSink.putShort((short) 0);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        verifyHeader(16, count, version, maxPayloadSize);
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws ShortPayloadException
     *             if the message's payload size is less than 16 bytes and
     *             <code>headerOnly</code> is <code>false</code>.
     */
    protected static ChannelAccessSubscriptionClientMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_EVENT_ADD
                .getCommandNumber());
        short mask;
        if (!headerOnly) {
            int payloadSize = messageHeader.getPayloadSize();
            if (payloadSize < EXPECTED_PAYLOAD_SIZE) {
                throw new ShortPayloadException(
                        "Received a CA_PROTO_EVENT_ADD message with a payload size of "
                                + payloadSize
                                + ". However, such a message should have a payload of at least "
                                + EXPECTED_PAYLOAD_SIZE + " bytes.");
            }
            int extraPayloadSize = payloadSize - EXPECTED_PAYLOAD_SIZE;
            byteSource.skip(12);
            mask = byteSource.getShort();
            byteSource.skip(2 + extraPayloadSize);
        } else {
            mask = 0;
        }
        return new ChannelAccessSubscriptionClientMessage(messageHeader,
                messageHeader.getDataType(), messageHeader.getCount(),
                messageHeader.getCID(), messageHeader.getContextSpecific(),
                mask);
    }

}
