/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

/**
 * <p>
 * Identity semantics for any reference. This wrapper can be used when identity
 * semantics are needed for objects which might override
 * {@link Object#equals(Object) equals(...)} and {@link Object#hashCode()
 * hashCode()}.
 * </p>
 *
 * <p>
 * This wrapper implements {@link #equals(Object)} in a way that an identity
 * wrapper is only equal to an identity wrapper that refers to exactly the same
 * object or if both identity wrappers refer to <code>null</code>. The
 * {@link #hashCode()} method is implemented so that it returns the value of
 * {@link System#identityHashCode(Object)} calculated for the reference.
 * </p>
 *
 * @param <T>
 *            type of the reference that is wrapped.
 * 
 * @author Sebastian Marsching
 */
public final class IdentityWrapper<T> {

    private T reference;

    /**
     * Creates an identity wrapper that wraps the specified reference.
     * 
     * @param reference
     *            reference that shall be wrapped. May be <code>null</code>.
     */
    public IdentityWrapper(T reference) {
        this.reference = reference;
    }

    /**
     * Returns the reference that is wrapped by this object. May return
     * <code>null</code> if this object was initialized with a reference to
     * <code>null</code>.
     * 
     * @return wrapped reference or <code>null</code>.
     */
    public T get() {
        return reference;
    }

    /**
     * <p>
     * Returns a hash code value for this object. This method is supported for
     * the benefit of hash tables. The hash code is calculated by calling
     * {@link System#identityHashCode(Object)}, passing the reference hold by
     * this identity wrapper.
     * </p>
     * 
     * <p>
     * If two identity wrappers are equal (as indicated by the return value of
     * {@link #equals(Object)}), they also have the same hash code. However, the
     * fact that two identity wrappers have the same hash code does not imply
     * that they are equal.
     * </p>
     * 
     * <p>
     * If the object reference wrapped by this identity wrapper does not
     * override the <code>hashCode</code> method but uses the default
     * implementation ({@link Object#hashCode()}), this method returns the same
     * value as <code>this.get().hashCode()</code>.
     * </p>
     * 
     * @return hash code suitable for use by a hash table.
     */
    @Override
    public int hashCode() {
        return System.identityHashCode(reference);
    }

    /**
     * <p>
     * Tells whether the specified object represents an identity wrapper that
     * refers to the same reference object.
     * </p>
     * 
     * <p>
     * Two identity wrappers are considered equal if and only if their
     * references are identical. If <code>obj</code> is an instance of
     * <code>IdentityWrapper</code>, calling this method has the same effect as
     * using <code>this.get() == obj.get()</code>. If both identity wrappers
     * refer to <code>null</code>, they are also considered equal.
     * </p>
     * 
     * @param obj
     *            reference object with which this object is compared.
     * @return <code>true</code> if <code>obj</code> is an identity wrapper that
     *         refers to the same object (checking for object identity, not
     *         equality) as this identity wrapper, <code>false</code> otherwise.
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof IdentityWrapper)) {
            return false;
        }
        IdentityWrapper<?> other = (IdentityWrapper<?>) obj;
        return this.reference == other.reference;
    }

    /**
     * Returns a string representation of this object. The string representation
     * is the string representation returned by the reference's
     * <code>toString()</code> method or the string "null" if this identity
     * wrapper's reference is <code>null</code>.
     * 
     * @return string representation of this object.
     */
    @Override
    public String toString() {
        if (reference == null) {
            return "null";
        } else {
            return reference.toString();
        }
    }

    /**
     * Creates an identity wrapper that wraps the specified reference. This is
     * exactly the same as using {@link IdentityWrapper#IdentityWrapper(Object)
     * new IdentityWrapper&lt;T&gt;(reference)}.
     * 
     * @param reference
     *            reference that shall be wrapped. May be <code>null</code>.
     * @param <T>
     *            type of the reference that is wrapped by the identity wrapper.
     * @return identity wrapper wrapping <code>reference</code>.
     */
    public static <T> IdentityWrapper<T> wrap(T reference) {
        return new IdentityWrapper<T>(reference);
    }

}
