/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;

/**
 * <p>
 * Simple concurrent lock-free queue backed by a linked list.
 * </p>
 *
 * <p>
 * Internally, this queue uses a very simple one-directional linked list that is
 * updated using atomic operations. This way, the queue is both thread safe and
 * lock free.
 * </p>
 * 
 * <p>
 * The {@link #add(Object)}, {@link #peek()}, and {@link #removeIfMatch(Object)}
 * operations are O(1). However, due to the concurrent nature of this
 * implementation, internal operations might fail occasionally and are repeated
 * until they succeed. Therefore, the time needed for a specific operation is
 * non-deterministic.
 * </p>
 *
 * @param <E>
 *            element type.
 * 
 * @author Sebastian Marsching
 */
public class SimpleConcurrentQueue<E> {

    /**
     * Linked-list node.
     * 
     * @param <E>
     *            queue element type.
     */
    private static class Node<E> {

        /**
         * Atomic updater for the {@link #next} field.
         */
        @SuppressWarnings("rawtypes")
        private static final AtomicReferenceFieldUpdater<Node, Node> NEXT_UPDATER = AtomicReferenceFieldUpdater
                .newUpdater(Node.class, Node.class, "next");

        /**
         * Atomic updater for the {@link #value} field.
         */
        @SuppressWarnings("rawtypes")
        private static final AtomicReferenceFieldUpdater<Node, Object> VALUE_UPDATER = AtomicReferenceFieldUpdater
                .newUpdater(Node.class, Object.class, "value");

        /**
         * Reference to the next node in the linked list. <code>null</code> if
         * this is the last node (tail).
         */
        @SuppressWarnings("unused")
        private volatile Node<E> next;

        /**
         * Value stored in this node.
         */
        @SuppressWarnings("unused")
        private volatile E value;

        /**
         * Creates a new node pointing to the specified next node and storing
         * the specified value.
         * 
         * @param next
         *            next node in the linked list or <code>null</code> if this
         *            is the last node.
         * @param value
         *            value being stored in this node. <code>null</code> means
         *            that this node is empty and should be ignored.
         */
        public Node(Node<E> next, E value) {
            this.next = next;
            this.value = value;
        }

        /**
         * Atomically compares and updates the reference to the next node.
         * 
         * @param expect
         *            expected value.
         * @param update
         *            new value.
         * @return <code>true</code> if the expected value matched the current
         *         value and the reference to the next node has been updated.
         *         <code>false</code> if the expected value did not match and
         *         those the reference to the next node has not been changed.
         */
        public boolean compareAndSetNext(Node<E> expect, Node<E> update) {
            return NEXT_UPDATER.compareAndSet(this, expect, update);
        }

        /**
         * Atomically compares and updates the value stored in this node. This
         * method checks for object identity, not equality.
         * 
         * @param expect
         *            expected value.
         * @param update
         *            new value.
         * @return <code>true</code> if the expected value matched the current
         *         value and the value has been updated. <code>false</code> if
         *         the expected value did not match and the current value has
         *         not been changed.
         */
        public boolean compareAndSetValue(E expect, E update) {
            return VALUE_UPDATER.compareAndSet(this, expect, update);
        }

        /**
         * Returns the reference to the next node.
         * 
         * @return next node or <code>null</code> if this is the last node in
         *         the list.
         */
        @SuppressWarnings("unchecked")
        public Node<E> getNext() {
            return NEXT_UPDATER.get(this);
        }

        /**
         * Returns the value stored in this node.
         * 
         * @return value stored in this node or <code>null</code> if this node
         *         is empty.
         */
        @SuppressWarnings("unchecked")
        public E getValue() {
            return (E) VALUE_UPDATER.get(this);
        }

    }

    /**
     * Atomic updater for the {@link #head} field.
     */
    @SuppressWarnings("rawtypes")
    private static final AtomicReferenceFieldUpdater<SimpleConcurrentQueue, Node> HEAD_UPDATER = AtomicReferenceFieldUpdater
            .newUpdater(SimpleConcurrentQueue.class, Node.class, "head");

    /**
     * First node in the queue. This is the node storing the element to be
     * returned next.
     */
    private volatile Node<E> head = new Node<E>(null, null);

    /**
     * Last node in the queue. This node can be the same as the {@link #head}
     * node if the queue stores less than two elements.
     */
    private volatile Node<E> tail = head;

    /**
     * Creates an empty queue.
     */
    public SimpleConcurrentQueue() {
    }

    /**
     * Adds an element to the tail of the queue.
     * 
     * @param e
     *            element to be added. Must not be <code>null</code>.
     */
    public void add(E e) {
        if (e == null) {
            throw new NullPointerException(
                    "Null elements are not supported by this queue.");
        }
        Node<E> newNode = new Node<E>(null, e);
        // Try to update the tail's next field with the new node. When we
        // succeed, this makes the new node the new tail.
        Node<E> oldTail;
        do {
            oldTail = this.tail;
        } while (!oldTail.compareAndSetNext(null, newNode));
        // The new node now is the new tail. We can simply set the tail (instead
        // of a compare-and-set) because this is the only place where we modify
        // the tail and no other thread can reach this point before we have
        // updated the tail (the other thread will be stuck in the loop above.
        this.tail = newNode;
    }

    /**
     * Removes the head of the queue if it matches the specified element.
     * 
     * @param expect
     *            element expected at the head of the queue.
     * @return <code>true</code> if the expected element was at the head of the
     *         queue and has been removed, <code>false</code> if the head of the
     *         queue did not match the expected element and thus was not
     *         removed.
     */
    public boolean removeIfMatch(E expect) {
        if (expect == null) {
            throw new NullPointerException(
                    "Null elements are not supported by this queue.");
        }
        Node<E> head = getHead();
        Node<E> node = head;
        do {
            E nodeValue = node.getValue();
            if (nodeValue == null) {
                // Another thread already removed the item or the head item is
                // not in the head node. We have to keep looking until we find
                // the item we are looking for, find a different item, or reach
                // the end of the queue.
                node = node.getNext();
            } else if (!nodeValue.equals(expect)) {
                // Another thread already removed the item and the corresponding
                // node.
                return false;
            } else {
                // The head is the node that we want to remove.
                if (node.compareAndSetValue(nodeValue, null)) {
                    // We successfully removed the value. Now, we can remove the
                    // head node.
                    Node<E> nextHead = node.getNext();
                    if (nextHead == null) {
                        nextHead = node;
                    }
                    if (nextHead != head) {
                        // We do not care about the result of the
                        // compare-and-set operation. If the head was already
                        // replaced by a different thread, that's fine.
                        compareAndSetHead(head, nextHead);
                    }
                    return true;
                } else {
                    // The compare-and-set was not successful. This means that
                    // another thread removed the element.
                    return false;
                }
            }
        } while (node != null);
        // We have reached the end of the queue without finding the element (or
        // any other element). This means that the queue is empty.
        return false;
    }

    /**
     * Returns the element at the head of the queue without removing it.
     * 
     * @return element at the head of the queue or <code>null</code> if the
     *         queue is empty.
     */
    public E peek() {
        Node<E> head = getHead();
        Node<E> node = head;
        do {
            E value = node.getValue();
            if (value != null) {
                // We found the first value.
                if (node != head) {
                    // If we skipped nodes, we make the current node the new
                    // head so that the next peek is quicker. We do not care
                    // whether the compare-and-set fails, this just means that
                    // another thread was quicker.
                    compareAndSetHead(head, node);
                }
                return value;
            } else {
                Node<E> nextNode = node.getNext();
                if (nextNode == null && node != head) {
                    // If we skipped nodes, we make the current node the new
                    // head so that the next peek is quicker. We do not care
                    // whether the compare-and-set fails, this just means that
                    // another thread was quicker.
                    compareAndSetHead(head, node);
                }
                node = nextNode;
            }
        } while (node != null);
        // We reached the end of the queue without finding a non-null value.
        // This means that the queue is empty.
        return null;
    }

    /**
     * Atomically compares and updates the reference to the head of the linked
     * list.
     * 
     * @param expect
     *            expected node.
     * @param update
     *            new head node.
     * @return <code>true</code> if the expected node was the current head node
     *         and has been replaced with the new node, <code>false</code> if
     *         the expected node did not match and thus the head was not
     *         changed.
     */
    private boolean compareAndSetHead(Node<E> expect, Node<E> update) {
        return HEAD_UPDATER.compareAndSet(this, expect, update);
    }

    /**
     * Returns the current head node.
     * 
     * @return current head node.
     */
    @SuppressWarnings("unchecked")
    private Node<E> getHead() {
        return HEAD_UPDATER.get(this);
    }

}
