/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.math.BigInteger;

/**
 * <p>
 * Time-stamp in the format used by Channel Access. A Channel Access time-stamp
 * has nano-second precision and uses a 32-bit unsigned integer for storing the
 * seconds since January 1st, 1990, 00:00:00 UTC and a 32-bit unsigned integer
 * for the nanoseconds part.
 * </p>
 * 
 * <p>
 * For obvious reasons, the nanoseconds part should only have values in the
 * range of 0 to 999999999. This implementation silently enforces those limits
 * by truncating values that are out of this range.
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public interface ChannelAccessTimeStamp extends
        Comparable<ChannelAccessTimeStamp> {

    /**
     * <p>
     * Returns the seconds since EPICS epoch. The EPICS epoch is January 1st,
     * 1990, 00:00:00 UTC. EPICS uses unsigned integers, so a negative value for
     * <code>seconds</code> actually indicates a value greater than the
     * {@link Integer#MAX_VALUE maximum positive integer value}.
     * </p>
     * 
     * @return seconds since January 1st, 1990, 00:00:00 UTC.
     */
    int getSeconds();

    /**
     * Returns the nanoseconds part. The nanoseconds part should be added to the
     * seconds part in order to get a more precise time-stamp. The nanoseconds
     * part is always a value in the range of 0 to 999999999.
     * 
     * @return nanoseconds nanoseconds part of the time-stamp.
     */
    int getNanoseconds();

    /**
     * Returns the time-stamp as the number of nanoseconds since EPICS epoch.
     * The EPICS epoch is January 1st, 1990, 00:00:00 UTC. This value is never
     * negative.
     * 
     * @return nanoseconds since January 1st, 1990, 00:00:00 UTC.
     */
    BigInteger toNanosecondsEPICS();

    /**
     * Returns the time-stamp as the number of nanoseconds since UNIX epoch. The
     * UNIX epoch is January 1st, 1970, 00:00:00 UTC. This value is never
     * negative.
     * 
     * @return nanoseconds since January 1st, 1970, 00:00:00 UTC.
     */
    BigInteger toNanosecondsUNIX();

    /**
     * Returns the time-stamp as a Java time-stamp, truncating the precision to
     * milliseconds. A Java time-stamp specified the number of milliseconds
     * since January 1st, 1970, 00:00:00 UTC.
     * 
     * @return milliseconds since January 1st, 1970, 00:00:00 UTC.
     */
    long toJavaTimeStamp();

}
