/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.DoubleBuffer;
import java.nio.charset.Charset;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeStamp;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Base class for Channel Access double type. This class contains the various
 * specializations of the double type as inlined classes.
 * 
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessDoubleBase<ImplementationType extends ChannelAccessDoubleBase<ImplementationType>>
        extends ChannelAccessValueBase<Double, ImplementationType> implements
        ChannelAccessDouble {

    /**
     * Implementation of DBR_DOUBLE.
     */
    public static final class ChannelAccessDoubleImpl extends
            ChannelAccessDoubleBase<ChannelAccessDoubleImpl> implements
            ChannelAccessSimpleOnlyDouble {

        /**
         * Creates a DBR_DOUBLE.
         * 
         * @param value
         *            double value.
         */
        public ChannelAccessDoubleImpl(double[] value) {
            super(value);
        }

        /**
         * Creates a read-only DBR_DOUBLE. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessDoubleImpl(ChannelAccessDoubleImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessDoubleImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_DOUBLE;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessSimpleOnlyDouble asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessDoubleImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessSimpleOnlyDouble clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessSimpleOnlyDouble) super.clone();
        }

        /**
         * Creates a DBR_DOUBLE from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_DOUBLE instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessDoubleImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessDoubleImpl value = new ChannelAccessDoubleImpl();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for double with alarms.
     */
    public static abstract class ChannelAccessAlarmDoubleBase<ImplementationType extends ChannelAccessAlarmDoubleBase<ImplementationType>>
            extends ChannelAccessDoubleBase<ImplementationType> implements
            ChannelAccessAlarmDouble {

        /**
         * Alarm status.
         */
        protected ChannelAccessAlarmStatus alarmStatus;

        /**
         * Alarm severity.
         */
        protected ChannelAccessAlarmSeverity alarmSeverity;

        /**
         * Base constructor.
         * 
         * @param value
         *            double value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        private ChannelAccessAlarmDoubleBase(double[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value);
            this.alarmStatus = alarmStatus;
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmDoubleBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmDoubleBase() {
        }

        @Override
        public ChannelAccessAlarmStatus getAlarmStatus() {
            if (delegate != null) {
                return delegate.getAlarmStatus();
            }
            return this.alarmStatus;
        }

        @Override
        public void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmStatus != null);
            this.alarmStatus = alarmStatus;
        }

        @Override
        public ChannelAccessAlarmSeverity getAlarmSeverity() {
            if (delegate != null) {
                return delegate.getAlarmSeverity();
            }
            return this.alarmSeverity;
        }

        @Override
        public void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmSeverity != null);
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Deserializes the alarm status from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeAlarms(ByteSource byteSource) {
            this.alarmStatus = ValueCodecUtils
                    .deserializeAlarmStatus(byteSource);
            this.alarmSeverity = ValueCodecUtils
                    .deserializeAlarmSeverity(byteSource);
        }

        /**
         * Serializes the alarm status to the byte sink.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeAlarms(ByteSink byteSink) {
            ValueCodecUtils.serializeAlarms(byteSink, alarmStatus,
                    alarmSeverity);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessAlarmDoubleBase<ImplementationType> other = (ChannelAccessAlarmDoubleBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.alarmStatus, other.alarmStatus)
                    .append(this.alarmSeverity, other.alarmSeverity).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.alarmStatus).append(this.alarmSeverity)
                    .toHashCode();
        }

        @Override
        public ChannelAccessAlarmDouble clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return (ChannelAccessAlarmDouble) super.clone();
        }

    }

    /**
     * Implementation of DBR_STS_DOUBLE.
     */
    public static final class ChannelAccessAlarmDoubleImpl extends
            ChannelAccessAlarmDoubleBase<ChannelAccessAlarmDoubleImpl>
            implements ChannelAccessAlarmOnlyDouble {

        /**
         * Creates a DBR_STS_DOUBLE.
         * 
         * @param value
         *            double value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         */
        public ChannelAccessAlarmDoubleImpl(double[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity) {
            super(value, alarmStatus, alarmSeverity);
        }

        /**
         * Creates a read-only DBR_STS_DOUBLE. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmDoubleImpl(
                ChannelAccessAlarmDoubleImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessAlarmDoubleImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_STS_DOUBLE;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            // Insert four extra padding bytes.
            byteSink.putInt(0);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessAlarmOnlyDouble asReadOnlyValue() {
            if (delegate != null) {
                return this;
            } else {
                return new ChannelAccessAlarmDoubleImpl(this);
            }
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessAlarmOnlyDouble clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessAlarmOnlyDouble) super.clone();
        }

        /**
         * Creates a DBR_STS_DOUBLE from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_STS_DOUBLE instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessAlarmDoubleImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessAlarmDoubleImpl value = new ChannelAccessAlarmDoubleImpl();
            value.deserializeAlarms(byteSource);
            // Skip four extra padding bytes.
            byteSource.skip(4);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_TIME_DOUBLE.
     */
    public static final class ChannelAccessTimeDoubleImpl extends
            ChannelAccessAlarmDoubleBase<ChannelAccessTimeDoubleImpl> implements
            ChannelAccessTimeDouble {

        /**
         * Seconds part of time-stamp.
         */
        protected int timeStampSeconds;

        /**
         * Nanoseconds part of time-stamp.
         */
        protected int timeStampNanoseconds;

        /**
         * Creates a DBR_TIME_DOUBLE.
         * 
         * @param value
         *            double value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param timeStampSeconds
         *            seconds part of time-stamp.
         * @param timeStampNanoSeconds
         *            nanoseconds part of time-stamp.
         */
        public ChannelAccessTimeDoubleImpl(double[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
                int timeStampNanoSeconds) {
            super(value, alarmStatus, alarmSeverity);
            this.timeStampSeconds = timeStampSeconds;
            if (timeStampNanoSeconds < 0 || timeStampNanoSeconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of timestamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoSeconds;
        }

        /**
         * Creates a read-only DBR_TIME_DOUBLE. All read operations are
         * delegated to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessTimeDoubleImpl(ChannelAccessTimeDoubleImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         */
        private ChannelAccessTimeDoubleImpl() {
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_TIME_DOUBLE;
        }

        @Override
        public ChannelAccessTimeStamp getTimeStamp() {
            if (delegate != null) {
                return delegate.getTimeStamp();
            }
            return ChannelAccessValueFactory.createTimeStamp(timeStampSeconds,
                    timeStampNanoseconds);
        }

        @Override
        public int getTimeSeconds() {
            if (delegate != null) {
                return delegate.getTimeSeconds();
            }
            return timeStampSeconds;
        }

        @Override
        public int getTimeNanoseconds() {
            if (delegate != null) {
                return delegate.getTimeNanoseconds();
            }
            return timeStampNanoseconds;
        }

        @Override
        public void setTimeStamp(ChannelAccessTimeStamp timeStamp) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStamp.getSeconds();
            this.timeStampNanoseconds = timeStamp.getNanoseconds();
        }

        @Override
        public void setTimeSeconds(int timeStampSeconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStampSeconds;
        }

        @Override
        public void setTimeNanoseconds(int timeStampNanoseconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (timeStampNanoseconds < 0 || timeStampNanoseconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of time-stamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoseconds;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            ValueCodecUtils.serializeTime(byteSink, timeStampSeconds,
                    timeStampNanoseconds);
            // Insert four extra padding bytes.
            byteSink.putInt(0);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessTimeDouble asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessTimeDoubleImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessTimeDoubleImpl other = (ChannelAccessTimeDoubleImpl) obj;
            return new EqualsBuilder()
                    .append(this.timeStampSeconds, other.timeStampSeconds)
                    .append(this.timeStampNanoseconds,
                            other.timeStampNanoseconds).isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.timeStampSeconds)
                    .append(this.timeStampNanoseconds).toHashCode();
        }

        @Override
        public ChannelAccessTimeDouble clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessTimeDouble) super.clone();
        }

        /**
         * Creates a DBR_TIME_DOUBLE from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @return DBR_TIME_DOUBLE instance created from the serialized data
         *         read from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessTimeDoubleImpl deserialize(
                ByteSource byteSource, int count) {
            ChannelAccessTimeDoubleImpl value = new ChannelAccessTimeDoubleImpl();
            value.deserializeAlarms(byteSource);
            // Deserialize time-stamp.
            value.timeStampSeconds = ValueCodecUtils
                    .deserializeTimeSeconds(byteSource);
            value.timeStampNanoseconds = ValueCodecUtils
                    .deserializeTimeNanoseconds(byteSource);
            // Skip four extra padding bytes.
            byteSource.skip(4);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for double with display information.
     */
    public static abstract class ChannelAccessGraphicsDoubleBase<ImplementationType extends ChannelAccessGraphicsDoubleBase<ImplementationType>>
            extends ChannelAccessAlarmDoubleBase<ImplementationType> implements
            ChannelAccessGraphicsDouble {

        private static final int ENGINEERING_UNITS_STRING_SIZE = 8;

        /**
         * Engineering units.
         */
        protected String engineeringUnits;

        /**
         * Raw representation of engineering units.
         */
        protected byte[] rawEngineeringUnits;

        /**
         * Upper display limit.
         */
        protected double upperDisplayLimit;

        /**
         * Lower display limit.
         */
        protected double lowerDisplayLimit;

        /**
         * Upper alarm limit.
         */
        protected double upperAlarmLimit;

        /**
         * Upper warning limit.
         */
        protected double upperWarningLimit;

        /**
         * Lower warning limit.
         */
        protected double lowerWarningLimit;

        /**
         * Lower alarm limit.
         */
        protected double lowerAlarmLimit;

        /**
         * Display precision.
         */
        protected short precision;

        /**
         * Charset used for encoding/decoding strings.
         */
        protected Charset charset;

        /**
         * Base constructor.
         * 
         * @param value
         *            double value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param precision
         *            display precision.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsDoubleBase(double[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, short precision,
                String engineeringUnits, double upperDisplayLimit,
                double lowerDisplayLimit, double upperAlarmLimit,
                double upperWarningLimit, double lowerWarningLimit,
                double lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity);
            this.precision = precision;
            this.engineeringUnits = engineeringUnits;
            assert (charset != null);
            this.charset = charset;
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
            this.upperDisplayLimit = upperDisplayLimit;
            this.lowerDisplayLimit = lowerDisplayLimit;
            this.upperAlarmLimit = upperAlarmLimit;
            this.upperWarningLimit = upperWarningLimit;
            this.lowerWarningLimit = lowerWarningLimit;
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsDoubleBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsDoubleBase(Charset charset) {
            this.charset = charset;
        }

        @Override
        public String getUnits() {
            if (delegate != null) {
                return delegate.getUnits();
            }
            return engineeringUnits;
        }

        @Override
        public byte[] getRawUnits() {
            if (delegate != null) {
                return delegate.getRawUnits();
            }
            return rawEngineeringUnits.clone();
        }

        @Override
        public void setUnits(String units) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.engineeringUnits = units;
            // Update raw representation.
            this.rawEngineeringUnits = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedFixedBytes(
                            this.engineeringUnits,
                            ENGINEERING_UNITS_STRING_SIZE, charset);
        }

        @Override
        public void setRawUnits(byte[] rawUnits) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            // We use a new array for storing the raw units. This way, we do
            // not have to copy the array in the clone() method and we do not
            // have to fill the array with null bytes explicitly.
            this.rawEngineeringUnits = new byte[ENGINEERING_UNITS_STRING_SIZE];
            // Copy first seven bytes.
            int copyLength = Math.min(rawUnits.length,
                    ENGINEERING_UNITS_STRING_SIZE - 1);
            System.arraycopy(rawUnits, 0, this.rawEngineeringUnits, 0,
                    copyLength);
            // Update string.
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
        }

        @Override
        public Double getGenericUpperDisplayLimit() {
            return getUpperDisplayLimit();
        }

        @Override
        public Double getGenericLowerDisplayLimit() {
            return getLowerDisplayLimit();
        }

        @Override
        public Double getGenericUpperAlarmLimit() {
            return getUpperAlarmLimit();
        }

        @Override
        public Double getGenericUpperWarningLimit() {
            return getUpperWarningLimit();
        }

        @Override
        public Double getGenericLowerWarningLimit() {
            return getLowerWarningLimit();
        }

        @Override
        public Double getGenericLowerAlarmLimit() {
            return getLowerAlarmLimit();
        }

        @Override
        public Charset getCharset() {
            if (delegate != null) {
                return delegate.getCharset();
            }
            return charset;
        }

        @Override
        public double getUpperDisplayLimit() {
            if (delegate != null) {
                return delegate.getUpperDisplayLimit();
            }
            return upperDisplayLimit;
        }

        @Override
        public void setUpperDisplayLimit(double upperDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperDisplayLimit = upperDisplayLimit;
        }

        @Override
        public double getLowerDisplayLimit() {
            if (delegate != null) {
                return delegate.getLowerDisplayLimit();
            }
            return lowerDisplayLimit;
        }

        @Override
        public void setLowerDisplayLimit(double lowerDisplayLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerDisplayLimit = lowerDisplayLimit;
        }

        @Override
        public double getUpperAlarmLimit() {
            if (delegate != null) {
                return delegate.getUpperAlarmLimit();
            }
            return upperAlarmLimit;
        }

        @Override
        public void setUpperAlarmLimit(double upperAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperAlarmLimit = upperAlarmLimit;
        }

        @Override
        public double getUpperWarningLimit() {
            if (delegate != null) {
                return delegate.getUpperWarningLimit();
            }
            return upperWarningLimit;
        }

        @Override
        public void setUpperWarningLimit(double upperWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperWarningLimit = upperWarningLimit;
        }

        @Override
        public double getLowerWarningLimit() {
            if (delegate != null) {
                return delegate.getLowerWarningLimit();
            }
            return lowerWarningLimit;
        }

        @Override
        public void setLowerWarningLimit(double lowerWarningLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerWarningLimit = lowerWarningLimit;
        }

        @Override
        public double getLowerAlarmLimit() {
            if (delegate != null) {
                return delegate.getLowerAlarmLimit();
            }
            return lowerAlarmLimit;
        }

        @Override
        public void setLowerAlarmLimit(double lowerAlarmLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerAlarmLimit = lowerAlarmLimit;
        }

        @Override
        public short getPrecision() {
            if (delegate != null) {
                return delegate.getPrecision();
            }
            return precision;
        }

        @Override
        public void setPrecision(short precision) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.precision = precision;
        }

        /**
         * Deserializes the display information from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeGraphics(ByteSource byteSource) {
            this.precision = byteSource.getShort();
            // Skip two extra padding bytes.
            byteSource.skip(2);
            this.rawEngineeringUnits = byteSource
                    .getByteArray(ENGINEERING_UNITS_STRING_SIZE);
            // Ensure that engineering units are null terminated.
            this.rawEngineeringUnits[ENGINEERING_UNITS_STRING_SIZE - 1] = 0;
            this.engineeringUnits = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(this.rawEngineeringUnits,
                            charset);
            this.upperDisplayLimit = byteSource.getDouble();
            this.lowerDisplayLimit = byteSource.getDouble();
            this.upperAlarmLimit = byteSource.getDouble();
            this.upperWarningLimit = byteSource.getDouble();
            this.lowerWarningLimit = byteSource.getDouble();
            this.lowerAlarmLimit = byteSource.getDouble();
        }

        /**
         * Serializes the display information.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeGraphics(ByteSink byteSink) {
            byteSink.putShort(precision);
            // Insert two extra padding bytes.
            byteSink.putShort((short) 0);
            byteSink.putByteArray(rawEngineeringUnits);
            byteSink.putDouble(upperDisplayLimit);
            byteSink.putDouble(lowerDisplayLimit);
            byteSink.putDouble(upperAlarmLimit);
            byteSink.putDouble(upperWarningLimit);
            byteSink.putDouble(lowerWarningLimit);
            byteSink.putDouble(lowerAlarmLimit);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessGraphicsDoubleBase<ImplementationType> other = (ChannelAccessGraphicsDoubleBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.precision, other.precision)
                    .append(this.engineeringUnits, other.engineeringUnits)
                    .append(this.rawEngineeringUnits, other.rawEngineeringUnits)
                    .append(this.upperDisplayLimit, other.upperDisplayLimit)
                    .append(this.lowerDisplayLimit, other.lowerDisplayLimit)
                    .append(this.upperAlarmLimit, other.upperAlarmLimit)
                    .append(this.upperWarningLimit, other.upperWarningLimit)
                    .append(this.lowerWarningLimit, other.lowerWarningLimit)
                    .append(this.lowerAlarmLimit, other.lowerAlarmLimit)
                    .append(this.charset, other.charset).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.precision).append(this.engineeringUnits)
                    .append(this.rawEngineeringUnits)
                    .append(this.upperDisplayLimit)
                    .append(this.lowerDisplayLimit)
                    .append(this.upperAlarmLimit)
                    .append(this.upperWarningLimit)
                    .append(this.lowerWarningLimit)
                    .append(this.lowerAlarmLimit).append(this.charset)
                    .toHashCode();
        }

        @Override
        public ChannelAccessGraphicsDouble clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            // We do not have to do anything special because we never modify
            // the rawEngineeringUnits array.
            return (ChannelAccessGraphicsDouble) super.clone();
        }

    }

    /**
     * Implementation of DBR_GR_DOUBLE.
     */
    public static final class ChannelAccessGraphicsDoubleImpl extends
            ChannelAccessGraphicsDoubleBase<ChannelAccessGraphicsDoubleImpl>
            implements ChannelAccessGraphicsOnlyDouble {

        /**
         * Creates a DBR_GR_DOUBLE.
         * 
         * @param value
         *            double value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param precision
         *            display precision.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessGraphicsDoubleImpl(double[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, short precision,
                String engineeringUnits, double upperDisplayLimit,
                double lowerDisplayLimit, double upperAlarmLimit,
                double upperWarningLimit, double lowerWarningLimit,
                double lowerAlarmLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, precision,
                    engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                    upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                    lowerAlarmLimit, charset);
        }

        /**
         * Creates a read-only DBR_GR_DOUBLE. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessGraphicsDoubleImpl(
                ChannelAccessGraphicsDoubleImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessGraphicsDoubleImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_GR_DOUBLE;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessGraphicsOnlyDouble asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessGraphicsDoubleImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessGraphicsOnlyDouble clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessGraphicsOnlyDouble) super.clone();
        }

        /**
         * Creates a DBR_GR_DOUBLE from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_GR_DOUBLE instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessGraphicsDoubleImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessGraphicsDoubleImpl value = new ChannelAccessGraphicsDoubleImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_CTRL_DOUBLE.
     */
    public static final class ChannelAccessControlsDoubleImpl extends
            ChannelAccessGraphicsDoubleBase<ChannelAccessControlsDoubleImpl>
            implements ChannelAccessControlsDouble {

        /**
         * Upper control limit.
         */
        protected double upperControlLimit;

        /**
         * Lower control limit.
         */
        protected double lowerControlLimit;

        /**
         * Creates a DBR_CTRL_DOUBLE.
         * 
         * @param value
         *            double value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param precision
         *            display precision.
         * @param engineeringUnits
         *            engineering units.
         * @param upperDisplayLimit
         *            upper display limit.
         * @param lowerDisplayLimit
         *            lower display limit.
         * @param upperAlarmLimit
         *            upper alarm limit.
         * @param upperWarningLimit
         *            upper warning limit.
         * @param lowerWarningLimit
         *            lower warning limit.
         * @param lowerAlarmLimit
         *            lower alarm limit.
         * @param upperControlLimit
         *            upper control limit.
         * @param lowerControlLimit
         *            lower control limit.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessControlsDoubleImpl(double[] value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, short precision,
                String engineeringUnits, double upperDisplayLimit,
                double lowerDisplayLimit, double upperAlarmLimit,
                double upperWarningLimit, double lowerWarningLimit,
                double lowerAlarmLimit, double upperControlLimit,
                double lowerControlLimit, Charset charset) {
            super(value, alarmStatus, alarmSeverity, precision,
                    engineeringUnits, upperDisplayLimit, lowerDisplayLimit,
                    upperAlarmLimit, upperWarningLimit, lowerWarningLimit,
                    lowerAlarmLimit, charset);
            this.upperControlLimit = upperControlLimit;
            this.lowerControlLimit = lowerControlLimit;
        }

        /**
         * Creates a read-only DBR_GR_DOUBLE. All read operations are delegated
         * to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessControlsDoubleImpl(
                ChannelAccessControlsDoubleImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessControlsDoubleImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_CTRL_DOUBLE;
        }

        @Override
        public Double getGenericUpperControlLimit() {
            return getUpperControlLimit();
        }

        @Override
        public Double getGenericLowerControlLimit() {
            return getLowerControlLimit();
        }

        @Override
        public double getUpperControlLimit() {
            if (delegate != null) {
                return delegate.getUpperControlLimit();
            }
            return upperControlLimit;
        }

        @Override
        public void setUpperControlLimit(double upperControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.upperControlLimit = upperControlLimit;
        }

        @Override
        public double getLowerControlLimit() {
            if (delegate != null) {
                return delegate.getLowerControlLimit();
            }
            return lowerControlLimit;
        }

        @Override
        public void setLowerControlLimit(double lowerControlLimit) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.lowerControlLimit = lowerControlLimit;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.length);
            serializeAlarms(byteSink);
            serializeGraphics(byteSink);
            byteSink.putDouble(upperControlLimit);
            byteSink.putDouble(lowerControlLimit);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessControlsDouble asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessControlsDoubleImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessControlsDoubleImpl other = (ChannelAccessControlsDoubleImpl) obj;
            return new EqualsBuilder()
                    .append(this.upperControlLimit, other.upperControlLimit)
                    .append(this.lowerControlLimit, other.lowerControlLimit)
                    .isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.upperControlLimit)
                    .append(this.lowerControlLimit).toHashCode();
        }

        @Override
        public ChannelAccessControlsDouble clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessControlsDouble) super.clone();
        }

        /**
         * Creates a DBR_CTRL_DOUBLE from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_CTRL_DOUBLE instance created from the serialized data
         *         read from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessControlsDoubleImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessControlsDoubleImpl value = new ChannelAccessControlsDoubleImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            value.deserializeGraphics(byteSource);
            value.upperControlLimit = byteSource.getDouble();
            value.lowerControlLimit = byteSource.getDouble();
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    protected double[] value;

    /**
     * Base constructor.
     * 
     * @param value
     *            double value.
     */
    private ChannelAccessDoubleBase(double[] value) {
        assert (value != null);
        this.value = value;
    }

    /**
     * Base constructor for read-only variant.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessDoubleBase(ImplementationType delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     */
    private ChannelAccessDoubleBase() {
    }

    @Override
    public int getValueSize() {
        if (delegate != null) {
            return delegate.getValueSize();
        }
        return value.length;
    }

    @Override
    public Double getGenericValueElement(int index) {
        if (delegate != null) {
            return delegate.getGenericValueElement(index);
        }
        return value[index];
    }

    @Override
    public DoubleBuffer getValue() {
        if (delegate != null) {
            return delegate.getValue().asReadOnlyBuffer();
        }
        return DoubleBuffer.wrap(value);
    }

    @Override
    public void setValue(double[] value) {
        if (delegate != null) {
            throw new UnsupportedOperationException(
                    "Cannot modify a read-only value.");
        }
        if (value == null) {
            throw new NullPointerException();
        }
        this.value = value;
    }

    /**
     * Deserializes the value elements from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @param count
     *            number of elements to deserialize.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    protected void deserializeValue(ByteSource byteSource, int count) {
        assert (count >= 0);
        if (count == 0) {
            byteSource.getDouble();
            value = ArrayUtils.EMPTY_DOUBLE_ARRAY;
        } else {
            value = byteSource.getDoubleArray(count);
        }
    }

    /**
     * Serializes the value elements to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write to.
     * @param count
     *            number of elements to be serialized.
     */
    protected void serializeValue(ByteSink byteSink, int count) {
        if (count == 0) {
            byteSink.putDouble(0.0);
        } else {
            byteSink.putDoubleArray(value, 0, count);
        }
    }

    @Override
    public boolean equals(Object obj) {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        if (!super.equals(obj)) {
            return false;
        }
        // Now, we can safely cast because the type has already been
        // checked.
        @SuppressWarnings("unchecked")
        ChannelAccessDoubleBase<ImplementationType> other = (ChannelAccessDoubleBase<ImplementationType>) obj;
        return new EqualsBuilder().append(this.value, other.value).isEquals();
    }

    @Override
    public int hashCode() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(value).toHashCode();
    }

    @Override
    public ChannelAccessDouble clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        @SuppressWarnings("unchecked")
        ChannelAccessDoubleBase<ImplementationType> copy = (ChannelAccessDoubleBase<ImplementationType>) super
                .clone();
        copy.value = this.value.clone();
        return copy;
    }

}
