/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;

/**
 * Utility functions for serializing and deserializing Channel Access values.
 * 
 * @author Sebastian Marsching
 */
public class ValueCodecUtils {

    /**
     * Private constructor. This class should never be instantiated.
     */
    private ValueCodecUtils() {
    }

    /**
     * Reads an alarm severity from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @return alarm severity read from the byte source.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    public static ChannelAccessAlarmStatus deserializeAlarmStatus(
            ByteSource byteSource) {
        return ChannelAccessAlarmStatus.forAlarmStatusCode(byteSource
                .getShort());
    }

    /**
     * Reads an alarm status from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @return alarm status read from the byte source.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    public static ChannelAccessAlarmSeverity deserializeAlarmSeverity(
            ByteSource byteSource) {
        try {
            return ChannelAccessAlarmSeverity.forSeverityCode(byteSource
                    .getShort());
        } catch (IllegalArgumentException e) {
            // We might have received an unknown severity from the network. We
            // treat this as the highest severity that we know of.
            return ChannelAccessAlarmSeverity.INVALID_ALARM;
        }
    }

    /**
     * Reads a boolean that is encoded as a short from the byte source. Any
     * non-zero number is interpreted as <code>true</code>, only zero is
     * interpreted as <code>false</code>.
     * 
     * @param byteSource
     *            byte source to read from.
     * @return boolean value read from the byte source.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    public static boolean deserializeBooleanFromShort(ByteSource byteSource) {
        return !(byteSource.getShort() == (short) 0);
    }

    /**
     * Reads the seconds part of a time-stamp from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @return seconds part of the time-stamp.
     */
    public static int deserializeTimeSeconds(ByteSource byteSource) {
        return byteSource.getInt();
    }

    /**
     * Reads the nanoseconds part of a time-stamp from the byte source.
     * 
     * @param byteSource
     *            byte source to read from.
     * @return nanoseconds part of the time-stamp.
     */
    public static int deserializeTimeNanoseconds(ByteSource byteSource) {
        return byteSource.getInt();
    }

    /**
     * Writes the alarm information to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write data to.
     * @param alarmStatus
     *            alarm status.
     * @param alarmSeverity
     *            alarm severity.
     * @throws BufferOverflowException
     *             if the byte sink throws such an exception.
     */
    public static void serializeAlarms(ByteSink byteSink,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity) {
        byteSink.putShort(alarmStatus.toAlarmStatusCode());
        byteSink.putShort(alarmSeverity.toSeverityCode());
    }

    /**
     * Writes the alarm severity to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write data to.
     * @param alarmSeverity
     *            alarm severity to be serialized.
     * @throws BufferOverflowException
     *             if the byte sink throws such an exception.
     */
    public static void serializeAlarmSeverity(ByteSink byteSink,
            ChannelAccessAlarmSeverity alarmSeverity) {
        byteSink.putShort(alarmSeverity.toSeverityCode());
    }

    /**
     * Writes the boolean to the byte sink by converting it to a short with the
     * value <code>1</code> (<code>true</code>) or 0 (<code>false</code>).
     * 
     * @param byteSink
     *            byte sink to write data to.
     * @param value
     *            boolean to be serialized.
     * @throws BufferOverflowException
     *             if the byte sink throws such an exception.
     */
    public static void serializeBooleanAsShort(ByteSink byteSink, boolean value) {
        byteSink.putShort(value ? ((short) 1) : ((short) 0));
    }

    /**
     * Writes the time-stamp to the byte sink.
     * 
     * @param byteSink
     *            byte sink to write data to.
     * @param seconds
     *            number of seconds.
     * @param nanoseconds
     *            number of nanoseconds.
     */
    public static void serializeTime(ByteSink byteSink, int seconds,
            int nanoseconds) {
        byteSink.putInt(seconds);
        byteSink.putInt(nanoseconds);
    }

}
