/*
 * Copyright 2018 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.nio.channels.spi.SelectorProvider;

import org.junit.Test;

import com.aquenos.epics.jackie.common.exception.NullErrorHandler;

/**
 * Tests for the {@link AbstractSocketChannelConnection}.
 * 
 * @author Sebastian Marsching
 */
public class AbstractSocketChannelConnectionTest {

    /**
     * Dummy implementation of the {@link CommunicationController} interface.
     * This implementation only implements the methods that we (indirectly) need
     * for our test of the {@link AbstractSocketChannelConnection}. Even the
     * methods that are implemented are limited to the functionality needed by
     * our tests, meaning that they may violate the contract specified by the
     * API.
     * 
     * @author Sebastian Marsching
     */
    private static class DummyCommunicationController
            implements CommunicationController {

        public DummyCommunicationController() throws IOException {
            this.selector = SelectorProvider.provider().openSelector();
        }

        private final Selector selector;

        @Override
        public SelectionKey registerChannel(ChannelProcessor channelProcessor,
                SelectableChannel channel) throws ClosedChannelException {
            // This method is actually called by the tested code and the
            // returned object is used, so we cannot return null. Unlike a real
            // implementation, we never use that registration in any way, so we
            // do not care about associating the channel processor with the
            // selection key.
            return channel.register(selector, 0);
        }

        @Override
        public void registerTimer(TimerProcessor timerProcessor,
                long eventTime) {
            // We do not use this, so we do not implement it.
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean inCommunicationThread() {
            // Our tests are single-threaded, so we can simply always return
            // true.
            return true;
        }

    }

    /**
     * <p>
     * Tests sending data through a connection's byte sink.
     * </p>
     * 
     * <p>
     * This test specifically tests for a bug that existed in released up to and
     * including EPICS Jackie 1.0.4. This bug lead to a
     * {@link IllegalArgumentException} being thrown by the {@code trySend()}
     * method, due to the incorrect use of
     * {@link ByteBufferByteSink#addBuffer(java.nio.ByteBuffer)} instead of
     * {@link ByteBufferByteSink#addBufferFirst(java.nio.ByteBuffer)}. It also
     * tests for another bug that was hidden by the first bug. This second bug
     * caused the number of bytes waiting to be sent to be tracked incorrectly
     * (the number would never get down to zero after a partially filled buffer
     * had been reused).
     * </p>
     * 
     * @throws IOException
     *             if there is an I/O error.
     * @throws InterruptedException
     *             if the thread running the tests is interrupted.
     */
    @Test
    public void sendData() throws IOException, InterruptedException {
        SocketChannel clientChannel = null;
        SocketChannel serverChannel = null;
        ServerSocketChannel serverMasterChannel = null;
        try {
            SelectorProvider selectorProvider = SelectorProvider.provider();
            DummyCommunicationController controller = new DummyCommunicationController();
            serverMasterChannel = selectorProvider.openServerSocketChannel();
            serverMasterChannel.socket().bind(null);
            clientChannel = selectorProvider.openSocketChannel();
            clientChannel.configureBlocking(false);
            clientChannel.connect(
                    serverMasterChannel.socket().getLocalSocketAddress());
            serverChannel = serverMasterChannel.accept();
            serverChannel.configureBlocking(false);
            while (!clientChannel.finishConnect()) {
                Thread.sleep(1L);
            }
            AbstractSocketChannelConnection connection = new AbstractSocketChannelConnection(
                    clientChannel, controller, new NullErrorHandler(), 1024, 0,
                    0) {
            };
            ByteSink sendSink = connection.getSendSink();
            int numberOfBytesSent = 1000000;
            byte nextByte = 0;
            for (int i = 0; i < numberOfBytesSent; ++i) {
                sendSink.putByte(nextByte);
                ++nextByte;
            }
            byte nextExpectedByte = 0;
            int numberOfBytesReceived = 0;
            ByteBuffer buffer = ByteBuffer.allocate(1024);
            while (connection.getSendQueueSizeInBytes() != 0) {
                assertNotEquals(numberOfBytesSent, numberOfBytesReceived);
                connection.trySend();
                serverChannel.read(buffer);
                buffer.flip();
                while (buffer.hasRemaining()) {
                    assertEquals(nextExpectedByte, buffer.get());
                    ++nextExpectedByte;
                    ++numberOfBytesReceived;
                }
                buffer.clear();
                Thread.sleep(1L);
            }
            while (numberOfBytesReceived != numberOfBytesSent) {
                serverChannel.read(buffer);
                buffer.flip();
                if (!buffer.hasRemaining()) {
                    Thread.sleep(1L);
                }
                while (buffer.hasRemaining()) {
                    assertEquals(nextExpectedByte, buffer.get());
                    ++nextExpectedByte;
                    ++numberOfBytesReceived;
                }
                buffer.clear();
            }
        } finally {
            if (clientChannel != null) {
                clientChannel.close();
            }
            if (serverChannel != null) {
                serverChannel.close();
            }
            if (serverMasterChannel != null) {
                serverMasterChannel.close();
            }
        }
    }

}
